#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import re
import os

from mlonmcu.models.model import ModelFormats


class TensorInfo:
    def __init__(self, name, shape, dtype, fix_names=False):
        self.name = name
        if fix_names:
            self.name = self.name.replace("/", "_").replace(";", "_")
        assert isinstance(shape, (tuple, list))
        self.shape = shape
        size_lookup = {
            "float32": 4,
            "uint8": 1,
            "int8": 1,
            "int32": 4,
            "int64": 8,
        }
        assert dtype in size_lookup, f"Unsupported type: {dtype}"
        self.dtype = dtype
        self.type_size = size_lookup[self.dtype]

    @property
    def size(self):
        ret = self.type_size
        for dim in self.shape:
            if isinstance(dim, complex):
                real = dim.real
                imag = dim.imag
                assert real == int(real)
                assert imag == int(imag)
                ret *= int(real) + int(imag)
            else:
                ret *= dim
        return ret


class TfLiteTensorInfo(TensorInfo):
    def __init__(self, t, fix_names=False):
        # Local imports to get rid of tflite dependency for non-tflite models
        from tflite.TensorType import TensorType as TType

        name = t.Name().decode()
        shape = tuple([t.Shape(si) for si in range(0, t.ShapeLength())])

        type_lookup = {
            TType.FLOAT32: "float32",
            TType.UINT8: "uint8",
            TType.INT8: "int8",
            TType.INT32: "int32",
            TType.BOOL: "int8",
            TType.INT64: "int64",
        }
        dtype = type_lookup[t.Type()]
        super().__init__(name, shape, dtype, fix_names=fix_names)


class RelayTensorInfo(TensorInfo):
    def __init__(self, name, shape, dtype, fix_names=False):
        super().__init__(name, shape, dtype, fix_names=fix_names)


class ModelInfo:
    def __init__(self, in_tensors, out_tensors, fix_names=False):
        self.in_tensors = in_tensors
        self.out_tensors = out_tensors

    def validate(self):
        assert len(self.in_tensors) > 0, "Missing inputs"
        assert len(self.out_tensors) > 0, "Missing outputs"

    @property
    def has_ins(self):
        return len(self.in_tensors) > 0

    @property
    def has_outs(self):
        return len(self.out_tensors) > 0


class TfLiteModelInfo(ModelInfo):
    def __init__(self, model, fix_names=False):
        assert model.SubgraphsLength() == 1
        g = model.Subgraphs(0)

        in_tensors = []
        for i in range(0, g.InputsLength()):
            t = g.Tensors(g.Inputs(i))
            in_tensors.append(TfLiteTensorInfo(t, fix_names=fix_names))

        out_tensors = []
        for i in range(0, g.OutputsLength()):
            t = g.Tensors(g.Outputs(i))
            out_tensors.append(TfLiteTensorInfo(t, fix_names=fix_names))
        super().__init__(in_tensors, out_tensors)


def shape_from_str(shape_str):
    return tuple(
        [complex(*map(int, x.split("i", 1))) if "i" in x else int(x) for x in shape_str.replace(" ", "").split(",")]
    )


def parse_relay_main(line):
    input_tensors = []
    output_tensors = []

    input_tensors_strs = re.compile(r"%[a-zA-Z0-9_]+\s?: Tensor\[\((?:\d+)(?:,\s*\d+)*\), (?:[a-zA-Z0-9_]+)\]").findall(
        line
    )
    for input_tensors_str in input_tensors_strs:
        res = re.compile(r"%([a-zA-Z0-9_]+)\s?: Tensor\[\(([\di]+(?:, [\di]+)*)\), ([a-zA-Z0-9_]+)\]").match(
            input_tensors_str
        )
        assert res is not None
        groups = res.groups()
        assert len(groups) == 3
        input_name, input_shape_str, input_type = groups
        if "v_param" in input_name:
            continue
        input_shape = shape_from_str(input_shape_str)
        input_tensor = TensorInfo(input_name, input_shape, input_type)
        input_tensors.append(input_tensor)

    output_tensor_names_str = re.compile(r"output_tensor_names=\[(\".*\")\]").findall(line)

    output_tensors_str = re.compile(r"-> (.+) {").findall(line)
    # The following depends on InferType annocations
    if len(output_tensors_str) > 0:
        output_tensor_strs = re.compile(r"Tensor\[\([\di]+(?:, [\di]+)*\), [a-zA-Z0-9_]+\]|(?:u?int\d+)").findall(
            output_tensors_str[0]
        )

        if len(output_tensor_names_str) > 0:
            output_tensor_names = re.compile(r"\"([a-zA-Z0-9_]+)\"").findall(output_tensor_names_str[0])
        else:
            output_tensor_names = [f"output{i}" for i in range(len(output_tensor_strs))]

        assert len(output_tensor_names) == len(output_tensor_strs)

        for i, output_name in enumerate(output_tensor_names):
            res = re.compile(r"Tensor\[\(([\di]+(?:, [\di]+)*)\), ([a-zA-Z0-9_]+)\]").match(output_tensor_strs[i])
            if res is None:
                res = re.compile(r"(u?int\d+)").match(output_tensor_strs[i])
            assert res is not None
            groups = res.groups()
            assert len(groups) in [1, 2]
            if len(groups) == 2:
                output_shape_str, output_type = groups
            elif len(groups) == 1:
                output_shape_str, output_type = "1, 1", groups[0]
            output_shape = shape_from_str(output_shape_str)
            output_tensor = TensorInfo(output_name, output_shape, output_type)
            output_tensors.append(output_tensor)
    return input_tensors, output_tensors


class RelayModelInfo(ModelInfo):
    def __init__(self, mod_text, fix_names=False):
        in_tensors = None
        out_tensors = None
        for line in mod_text.split("\n"):
            if "def @main(" in line:
                in_tensors, out_tensors = parse_relay_main(line)
                break
        assert (
            in_tensors is not None and len(in_tensors) > 0
        ), "RelayModelInfo: input_tensors not found (Add TypeInfer details or provided types/shapes manually)"
        assert (
            out_tensors is not None and len(out_tensors) > 0
        ), "RelayModelInfo: output tensors not found (Add TypeInfer details or provide types/shapes manually)"
        super().__init__(in_tensors, out_tensors)


def get_tfgraph_inout(graph):
    ops = graph.get_operations()
    outputs_set = set(ops)
    inputs = []
    for op in ops:
        if op.type == "Placeholder":
            inputs.append(op)
        else:
            for input_tensor in op.inputs:
                if input_tensor.op in outputs_set:
                    outputs_set.remove(input_tensor.op)
    outputs = list(outputs_set)
    return (inputs, outputs)


class PBModelInfo(ModelInfo):
    def __init__(self, model_file):
        import tensorflow as tf

        with tf.io.gfile.GFile(model_file, "rb") as f:
            graph_def = tf.compat.v1.GraphDef()
            graph_def.ParseFromString(f.read())
        with tf.Graph().as_default() as graph:
            tf.import_graph_def(graph_def)
            inputs, outputs = get_tfgraph_inout(graph)
        in_tensors = [TensorInfo(t.name, t.shape.as_list(), t.dtype.name) for op in inputs for t in op.outputs]
        out_tensors = [TensorInfo(t.name, t.shape.as_list(), t.dtype.name) for op in outputs for t in op.outputs]
        super().__init__(in_tensors, out_tensors)


class ONNXModelInfo(ModelInfo):
    def __init__(self, model_file):
        from google.protobuf.json_format import MessageToDict
        import onnx
        from onnx.mapping import TENSOR_TYPE_TO_NP_TYPE

        model = onnx.load(model_file)

        def _helper(tensors):
            ret = []
            for tensor in tensors:
                d = MessageToDict(tensor)
                name = d["name"]
                tensor_type = d["type"]["tensorType"]
                elem_type = tensor_type["elemType"]
                dims = tensor_type["shape"]["dim"]
                shape = [int(x["dimValue"]) if "dimValue" in x else 40 for x in dims]  # TODO: dyn shape
                dtype = str(TENSOR_TYPE_TO_NP_TYPE[elem_type])
                ret.append(TensorInfo(name, shape, dtype))
            return ret

        in_tensors = _helper(model.graph.input)
        out_tensors = _helper(model.graph.output)

        # TVM seems to ignore the original output names for ONNX models
        if len(out_tensors) == 1:
            out_tensors[0].name = "output"
        else:
            for i, t in enumerate(out_tensors):
                t.name = f"output{i}"
        super().__init__(in_tensors, out_tensors)


class PaddleModelInfo(ModelInfo):
    def __init__(self, model_file):
        import paddle

        paddle.enable_static()
        paddle.disable_signal_handler()

        exe = paddle.static.Executor(paddle.CPUPlace())
        prefix = "".join(model_file.strip().split(".")[:-1])
        program, _, _ = paddle.static.load_inference_model(prefix, exe)

        output_names = list()
        for block in program.blocks:
            for op in block.ops:
                if op.type == "fetch":
                    output_names.append(op.input("X")[0])

        def _convert_dtype_value(val):
            """Converts a Paddle type id to a string."""
            # See: https://github.com/apache/tvm/blob/cc769fdc951707b8be991949864817b955a4dbc7/python/tvm/
            # relay/frontend/paddlepaddle.py#L70

            convert_dtype_map = {
                21: "int8",
                20: "uint8",
                6: "float64",
                5: "float32",
                4: "float16",
                3: "int64",
                2: "int32",
                1: "int16",
                0: "bool",
            }
            if val not in convert_dtype_map:
                msg = "Paddle data type value %d is not handled yet." % (val)
                raise NotImplementedError(msg)
            return convert_dtype_map[val]

        raise NotImplementedError
        # super().__init__(in_tensors, out_tensors)


def get_tflite_model_info(model_buf):
    # Local imports to get rid of tflite dependency for non-tflite models
    import tflite

    tflite_model = tflite.Model.GetRootAsModel(model_buf, 0)
    model_info = TfLiteModelInfo(tflite_model)
    return model_info


def get_relay_model_info(mod_text):
    model_info = RelayModelInfo(mod_text)
    return model_info


def get_pb_model_info(model_file):
    model_info = PBModelInfo(model_file)
    return model_info


def get_paddle_model_info(model_file):
    model_info = PaddleModelInfo(model_file)
    return model_info


def get_onnx_model_info(model_file):
    model_info = ONNXModelInfo(model_file)
    return model_info


def get_model_info(model, backend_name="unknown"):
    ext = os.path.splitext(model)[1][1:]
    fmt = ModelFormats.from_extension(ext)
    if fmt == ModelFormats.TFLITE:
        with open(model, "rb") as handle:
            model_buf = handle.read()
            return "tflite", get_tflite_model_info(model_buf)
    elif fmt == ModelFormats.RELAY:
        # Warning: the wrapper generateion does currently not work because of the
        # missing possibility to get the relay models input names and shapes
        with open(model, "r") as handle:
            mod_text = handle.read()
        return "relay", get_relay_model_info(mod_text)
    elif fmt == ModelFormats.PB:
        return "pb", get_pb_model_info(model)
    elif fmt == ModelFormats.ONNX:
        return "onnx", get_onnx_model_info(model)
    elif fmt == ModelFormats.PADDLE:
        return "pdmodel", get_pb_model_info(model)
    else:
        raise RuntimeError(f"Unsupported model format '{fmt.name}' for backend '{backend_name}'")


def get_fallback_model_info(model, input_shapes, output_shapes, input_types, output_types, backend_name="unknown"):
    ext = os.path.splitext(model)[1][1:]
    fmt = ModelFormats.from_extension(ext)

    def helper(shapes, types):
        return [TensorInfo(name, shape, types[name]) for name, shape in shapes.items()]

    info = ModelInfo(
        in_tensors=helper(input_shapes, input_types),
        out_tensors=helper(output_shapes, output_types),
    )

    if fmt == ModelFormats.TFLITE:
        return "tflite", info
    elif fmt == ModelFormats.RELAY:
        return "relay", info
    elif fmt == ModelFormats.PB:
        return "pb", info
    elif fmt == ModelFormats.ONNX:
        return "onnx", info
    elif fmt == ModelFormats.PADDLE:
        return "pdmodel", info
    else:
        raise RuntimeError(f"Unsupported model format '{fmt.name}' for backend '{backend_name}'")


def get_model_format(model):
    ext = os.path.splitext(model)[1][1:]
    fmt = ModelFormats.from_extension(ext)

    if fmt:
        return fmt.extension
    else:
        return ext


def get_supported_formats():
    return [ModelFormats.TFLITE, ModelFormats.RELAY, ModelFormats.PB, ModelFormats.ONNX, ModelFormats.PADDLE]



# from mlonmcu.artifact import ArtifactFormat
from mlonmcu.artifact import lookup_artifacts

from mlonmcu.testing.helpers import _test_backend, DEFAULT_TFLITE_MODELS

import pytest


def get_tvm_example_config(backend):
    return {
        f"{backend}.extra_pass_config": {"relay.FuseOps.max_depth": 0},  # TODO
        f"{backend}.disabled_passes": ["AlterOpLayout"],
        f"{backend}.target_device": "arm_cpu",
        f"{backend}.opt_level": 2,
    }


@pytest.mark.user_context
@pytest.mark.parametrize("model_name", DEFAULT_TFLITE_MODELS)
@pytest.mark.parametrize("feature_names", [[]])
@pytest.mark.parametrize(
    "config", [{}, {"tflmi.arena_size": 2**20, "tflmi.ops": ["TODO"]}]  # TODO
)  # TODO: user should be ablte to overwrite sesstings parsed by frontend
def test_backend_tflmi(user_context, model_name, models_dir, feature_names, config):
    df, artifacts = _test_backend("tflmi", user_context, model_name, models_dir, feature_names, config)
    assert df["Framework"][0] == "tflm"

    assert len(lookup_artifacts(artifacts, name="model.cc")) == 1
    assert len(lookup_artifacts(artifacts, name="model.cc.h")) == 1
    assert len(lookup_artifacts(artifacts, name="tflmi_arena_size.txt")) == 1
    # TODO: check if non-empty
    # TODO: check if arena size is propagated


@pytest.mark.user_context
@pytest.mark.parametrize("model_name", DEFAULT_TFLITE_MODELS)
@pytest.mark.parametrize("feature_names", [[]])
@pytest.mark.parametrize("config", [{}])
def test_backend_tflmc(user_context, model_name, models_dir, feature_names, config):
    df, artifacts = _test_backend("tflmc", user_context, model_name, models_dir, feature_names, config)
    assert df["Framework"][0] == "tflm"

    assert len(lookup_artifacts(artifacts, name="model.cc")) == 1
    assert len(lookup_artifacts(artifacts, name="model.cc.h")) == 1
    assert len(lookup_artifacts(artifacts, name="tflmc_out.log")) == 1


@pytest.mark.slow
@pytest.mark.user_context
# @pytest.mark.parametrize("model_name", DEFAULT_TFLITE_MODELS + DEFAULT_RELAY_MODELS
# + DEFAULT_PB_MODELS + DEFAULT_ONNX_MODELS + DEFAULT_PADDLE_MODELS)
@pytest.mark.parametrize("model_name", DEFAULT_TFLITE_MODELS)
@pytest.mark.parametrize("feature_names", [[]])
@pytest.mark.parametrize(
    "config",
    [
        {},
        {
            **get_tvm_example_config("tvmaot"),
            "tvmaot.arena_size": 2**20,
            "tvmaot.alignment_bytes": 16,
        },
    ],
)
def test_backend_tvmaot(user_context, model_name, models_dir, feature_names, config):
    df, artifacts = _test_backend("tvmaot", user_context, model_name, models_dir, feature_names, config)
    assert df["Framework"][0] == "tvm"

    assert len(lookup_artifacts(artifacts, name="default.tar")) == 1
    assert len(lookup_artifacts(artifacts, name="tvmc_compile_out.log")) == 1
    # TODO: check if non-empty
    # TODO: check if arena/alignment updated


@pytest.mark.slow
@pytest.mark.user_context
@pytest.mark.parametrize("model_name", DEFAULT_TFLITE_MODELS)
@pytest.mark.parametrize("feature_names", [[]])
@pytest.mark.parametrize(
    "config",
    [
        {},
        {
            **get_tvm_example_config("tvmaotplus"),
            "tvmaotplus.arena_size": 2**20,
            "tvmaotplus.alignment_bytes": 16,
        },
    ],
)
def test_backend_tvmaotplus(user_context, model_name, models_dir, feature_names, config):
    df, artifacts = _test_backend("tvmaotplus", user_context, model_name, models_dir, feature_names, config)
    assert df["Framework"][0] == "tvm"

    assert len(lookup_artifacts(artifacts, name="default.tar")) == 1
    assert len(lookup_artifacts(artifacts, name="tvmc_compile_out.log")) == 1
    # TODO: check if non-empty
    # TODO: check if arena/alignment updated


@pytest.mark.slow
@pytest.mark.user_context
@pytest.mark.parametrize("model_name", DEFAULT_TFLITE_MODELS)
@pytest.mark.parametrize("feature_names", [[]])
@pytest.mark.parametrize(
    "config",
    [
        {},
        {
            **get_tvm_example_config("tvmrt"),
            "tvmrt.arena_size": 2**20,
        },
    ],
)  # TODO: combine tvm common configs
def test_backend_tvmrt(user_context, model_name, models_dir, feature_names, config):
    df, artifacts = _test_backend("tvmrt", user_context, model_name, models_dir, feature_names, config)
    assert df["Framework"][0] == "tvm"

    assert len(lookup_artifacts(artifacts, name="default.tar")) == 1
    assert len(lookup_artifacts(artifacts, name="tvmc_compile_out.log")) == 1
    # TODO: check if non-empty
    # TODO: check if arena updated


@pytest.mark.slow
@pytest.mark.user_context
@pytest.mark.parametrize("model_name", DEFAULT_TFLITE_MODELS)
@pytest.mark.parametrize("feature_names", [[]])
@pytest.mark.parametrize(
    "config",
    [
        {},
        {
            **get_tvm_example_config("tvmcg"),
            "arena_size": 2**20,
        },
    ],
)
def test_backend_tvmcg(user_context, model_name, models_dir, feature_names, config):
    df, artifacts = _test_backend("tvmcg", user_context, model_name, models_dir, feature_names, config)
    assert df["Framework"][0] == "tvm"

    assert len(lookup_artifacts(artifacts, name="default.tar")) == 1
    assert len(lookup_artifacts(artifacts, name="tvmc_compile_out.log")) == 1
    # TODO: check if non-empty
    # TODO: check if arena updated



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""TVM Platform"""

from .tvm_base_platform import TvmBasePlatform
from .tvm_target_platform import TvmTargetPlatform
from .tvm_build_platform import TvmBuildPlatform
from .tvm_tune_platform import TvmTunePlatform


class TvmPlatform(
    TvmBasePlatform,
    TvmBuildPlatform,
    # TvmTargetPlatform,  # implicitly via tune platform
    TvmTunePlatform,
):
    """TVM Platform class."""

    FEATURES = (
        TvmBasePlatform.FEATURES | TvmBuildPlatform.FEATURES | TvmTargetPlatform.FEATURES | TvmTunePlatform.FEATURES
    )  # TODO: validate?

    DEFAULTS = {
        **TvmBasePlatform.DEFAULTS,
        **TvmBuildPlatform.DEFAULTS,
        **TvmTargetPlatform.DEFAULTS,
        **TvmTunePlatform.DEFAULTS,
    }

    REQUIRED = (
        TvmBasePlatform.REQUIRED | TvmBuildPlatform.REQUIRED | TvmTargetPlatform.REQUIRED | TvmTunePlatform.REQUIRED
    )

    def __init__(self, features=None, config=None):
        super(TvmPlatform, self).__init__(
            "tvm",
            features=features,
            config=config,
        )
        self.tempdir = None
        self.project_name = "app"
        self.project_dir = None



import pytest
from mlonmcu.environment.config import PathConfig
from mlonmcu.session.run import RunStage
from mlonmcu.artifact import lookup_artifacts

# TODO: make sure that we use quant/float models and several different operators

DEFAULT_TFLITE_MODELS = ["sine_model"]
DEFAULT_ONNX_MODELS = ["onnx_mnist"]
DEFAULT_RELAY_MODELS = ["test_cnn"]
DEFAULT_PB_MODELS = ["mobilenet_v1_1_0_224"]
ALL_PB_MODELS = ["mobilenet_v1_1_0_224", "mobilenet_v1_1_0_224_quant", "mobilenet_v1_0_25_128"]
DEFAULT_PADDLE_MODELS = ["paddle_resnet50"]

MODEL_FRONTENDS = {
    "sine_model": "tflite",
    "aww": "tflite",
    "vww": "tflite",
    "resnet": "tflite",
    "toycar": "tflite",
    "tinymlperf": "tflite",
    "onnx_mnist": "onnx",
    "test_cnn": "relay",
    "mobilenet_v1_1.0_224_frozen": "pb",  # or tflite
    "mobilenet_v1_1.0_224_quant": "pb",  # or tflite
    "mobilenet_v1_0.25_128": "pb",  # or tflite
    "paddle_resnet50": "paddle",
}

DEFAULT_FRONTENDS = ["tflite"]  # TODO: needs to match with the DEFAULT_MODELS
ALL_FRONTENDS = ["tflite", "pb", "relay", "onnx"]

TARGET_PLATFORMS = {
    "host_x86": "mlif",
    "corstone300": "mlif",
    "etiss": "mlif",
    "spike": "mlif",
    "ovpsim": "mlif",
    "riscv_qemu": "mlif",
    "tvm_cpu": "tvm",
    "microtvm_host": "microtvm",
    "microtvm_spike": "microtvm",
    "microtvm_espidf": "microtvm",
    "microtvm_etiss": "microtvm",
    "microtvm_zephyr": "microtvm",
    "microtvm_arduino": "microtvm",
}

DEFAULT_BACKENDS = ["tflmi", "tvmaot"]  # TODO: how about tvmrt/tflmc
ALL_BACKENDS = ["tflmi", "tflmc", "tvmaot", "tvmaotplus", "tvmrt", "tvmcg"]

# DEFAULT_MLIF_TARGETS = ["host_x86", "etiss", "spike", "ovpsim", "corstone300"]
DEFAULT_MLIF_TARGETS = ["host_x86", "etiss", "spike", "corstone300"]
ALL_MLIF_TARGETS = ["host_x86", "etiss", "spike", "ovpsim", "corstone300", "riscv_qemu"]

DEFAULT_ESPIDF_TARGETS = ["esp32", "esp32c3"]
# ALL_ESPIDF_TARGETS = [] # TODO
DEFAULT_TARGETS = DEFAULT_MLIF_TARGETS + DEFAULT_ESPIDF_TARGETS

VEXT_TARGETS = ["spike", "ovpsim", "riscv_qemu"]  # TODO: etiss
# Spike PEXT support (0.92) is outdated compared to ovpsim (0.96)
PEXT_TARGETS = ["ovpsim"]  # TODO: etiss, spike

# RISCV_TARGETS = ["spike", "etiss", "ovpsim"]
RISCV_TARGETS = ["spike", "etiss", "ovpsim", "riscv_qemu"]

ARM_MVEI_TARGETS = ["corstone300"]
ARM_DSP_TARGETS = ["corstone300"]

DEBUG_ARENA_BACKENDS = ["tflmi", "tvmaot", "tvmaotplus", "tvmrt"]


def _check_features(user_context, feature_names):
    for feature in feature_names:
        if not user_context.environment.has_feature(feature):
            pytest.skip(f"Feature '{feature}' is not enabled.")


def _init_run(user_context, models_dir, config):
    user_context.environment.paths["models"] = [PathConfig(models_dir)]
    session = user_context.create_session()
    return session, session.create_run(config=config)


def _test_frontend(frontend_name, user_context, model_name, models_dir, feature_names, config):
    user_config = user_context.environment.vars.copy()
    user_config.update(config)
    if not user_context.environment.has_frontend(frontend_name):
        pytest.skip(f"Frontend '{frontend_name}' is not enabled.")
    _check_features(user_context, feature_names)
    session, run = _init_run(user_context, models_dir, user_config)
    with session:
        run.add_features_by_name(feature_names, context=user_context)
        run.add_frontend_by_name(frontend_name, context=user_context)
        run.add_model_by_name(model_name, context=user_context)
        assert session.process_runs(until=RunStage.LOAD, context=user_context)
    report = session.get_reports()
    df, artifacts = report.df, run.artifacts

    assert len(df) == 1
    assert df["Model"][0] == model_name
    assert df["Frontend"][0] == frontend_name

    assert len(lookup_artifacts(artifacts)) > 0
    return df, artifacts

    # TODO: test for metadata
    # TODO: test model data.c (after refactor)
    # artifacts.append(Artifact(f"{name}.{ext}", raw=raw, fmt=ArtifactFormat.RAW))
    # data_artifact = Artifact("data.c", content=data_src, fmt=ArtifactFormat.SOURCE)


def _test_backend(backend_name, user_context, model_name, models_dir, feature_names, config):
    user_config = user_context.environment.vars.copy()
    user_config.update(config)
    frontend_name = MODEL_FRONTENDS[model_name]
    if not user_context.environment.has_frontend(frontend_name):
        pytest.skip(f"Frontend '{frontend_name}' is not enabled.")
    if not user_context.environment.has_backend(backend_name):
        pytest.skip(f"Backend '{backend_name}' is not enabled.")
    _check_features(user_context, feature_names)
    session, run = _init_run(user_context, models_dir, user_config)
    with session:
        run.add_features_by_name(feature_names, context=user_context)
        run.add_frontend_by_name(frontend_name, context=user_context)
        run.add_model_by_name(model_name, context=user_context)
        run.add_backend_by_name(backend_name, context=user_context)  # TODO: implicit Framework
        assert session.process_runs(until=RunStage.BUILD, context=user_context)
    report = session.get_reports()
    df, artifacts = report.df, run.artifacts

    assert len(df) == 1
    assert df["Model"][0] == model_name
    assert df["Backend"][0] == backend_name
    return df, artifacts


def _test_compile_platform(
    platform_name, backend_name, target_name, user_context, model_name, models_dir, feature_names, config
):
    user_config = user_context.environment.vars.copy()
    user_config.update(config)
    frontend_name = MODEL_FRONTENDS[model_name]
    if platform_name is None:
        platform_name = TARGET_PLATFORMS[target_name]
    if not user_context.environment.has_frontend(frontend_name):
        pytest.skip(f"Frontend '{frontend_name}' is not enabled.")
    if not user_context.environment.has_backend(backend_name):
        pytest.skip(f"Backend '{backend_name}' is not enabled.")
    if not user_context.environment.has_platform(platform_name):
        pytest.skip(f"Platform '{platform_name}' is not enabled.")  # TODO: not enabled -> not installed
    if not user_context.environment.has_target(target_name):
        pytest.skip(f"Target '{target_name}' is not enabled.")  # TODO: remove check?
    _check_features(user_context, feature_names)
    session, run = _init_run(user_context, models_dir, user_config)
    with session:
        run.add_features_by_name(feature_names, context=user_context)
        run.add_frontend_by_name(frontend_name, context=user_context)
        run.add_model_by_name(model_name, context=user_context)
        run.add_backend_by_name(backend_name, context=user_context)  # TODO: implicit Framework
        run.add_platform_by_name(platform_name, context=user_context)
        run.add_target_by_name(target_name, context=user_context)
        assert session.process_runs(until=RunStage.COMPILE, context=user_context)
    report = session.get_reports()
    df, artifacts = report.df, run.artifacts

    assert len(df) == 1
    assert df["Model"][0] == model_name
    assert df["Platform"][0] == platform_name
    assert df["Target"][0] == target_name
    return df, artifacts


def _test_run_platform(
    platform_name, backend_name, target_name, user_context, model_name, models_dir, feature_names, config
):
    user_config = user_context.environment.vars.copy()
    user_config.update(config)
    frontend_name = MODEL_FRONTENDS[model_name]
    if platform_name is None:
        platform_name = TARGET_PLATFORMS[target_name]
    if not user_context.environment.has_frontend(frontend_name):
        pytest.skip(f"Frontend '{frontend_name}' is not enabled.")
    if not user_context.environment.has_backend(backend_name):
        pytest.skip(f"Backend '{backend_name}' is not enabled.")
    if not user_context.environment.has_platform(platform_name):
        pytest.skip(f"Platform '{platform_name}' is not enabled.")  # TODO: not enabled -> not installed
    if not user_context.environment.has_target(target_name):
        pytest.skip(f"Target '{target_name}' is not enabled.")  # TODO: remove check?
    _check_features(user_context, feature_names)
    session, run = _init_run(user_context, models_dir, user_config)
    with session:
        run.add_features_by_name(feature_names, context=user_context)
        run.add_frontend_by_name(frontend_name, context=user_context)
        run.add_model_by_name(model_name, context=user_context)
        run.add_backend_by_name(backend_name, context=user_context)  # TODO: implicit Framework
        run.add_platform_by_name(platform_name, context=user_context)
        run.add_target_by_name(target_name, context=user_context)
        assert session.process_runs(until=RunStage.RUN, context=user_context)
    report = session.get_reports()
    df, artifacts = report.df, run.artifacts

    assert len(df) == 1
    assert df["Model"][0] == model_name
    assert df["Platform"][0] == platform_name
    assert df["Target"][0] == target_name
    return df, artifacts



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""UMA backend for the q_vanilla_accelerator accelerator"""
from .passes import QVanillaAcceleratorConv2dPass, ConvertLayout

# , ConvertLayout, Canonicalize, QVanillaAcceleratorDensePass
from tvm.relay.backend.contrib.uma.api.utils import PassPhase
from tvm.relay.backend.contrib.uma.backend import UMABackend
from .codegen import gen_includes

# from .patterns import conv2d_pattern, dense_pattern, qnn_conv2d_pattern
from .patterns import qnn_conv2d_add_pattern
from .strategies import qnn_conv2d_strategy


class QVanillaAcceleratorBackend(UMABackend):
    """UMA backend for the QVanillaAccelerator accelerator."""

    def __init__(self):
        print("init!")
        super().__init__()

        # Target configuration
        self._register_target_attr("dimension")

        # Relay Pattern registration
        self._register_pattern("qnn_conv2d_add", qnn_conv2d_add_pattern())

        # self._register_pattern("dense", dense_pattern())

        # Relay to Relay function registration
        self._register_relay_pass(PassPhase.PRE_PARTITIONING, ConvertLayout())

        # self._register_relay_pass(PassPhase.POST_PARTITIONING_0, Canonicalize())

        # Relay to TIR function registration
        self._register_operator_strategy("qnn.conv2d", qnn_conv2d_strategy)

        self._register_tir_pass(PassPhase.TIR_PHASE_0, QVanillaAcceleratorConv2dPass())

        # self._register_tir_pass(PassPhase.TIR_PHASE_0, QVanillaAcceleratorDensePass())

        # TIR to runtime function registration
        self._register_codegen(fmt="c", includes=gen_includes)

    @property
    def target_name(self):
        return "q_vanilla_accelerator"



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import time
import argparse
from pathlib import Path
from abc import ABC, abstractmethod
from typing import Tuple, List

from mlonmcu.cli.helper.parse import extract_feature_names, extract_config
from mlonmcu.feature.type import FeatureType
from mlonmcu.config import filter_config
from mlonmcu.feature.features import get_matching_features
from mlonmcu.artifact import Artifact, ArtifactFormat
from mlonmcu.logging import get_logger
from mlonmcu.target.metrics import Metrics

logger = get_logger()


class Backend(ABC):
    name = None

    FEATURES = set()
    DEFAULTS = {}
    REQUIRED = set()
    OPTIONAL = set()

    def __init__(
        self,
        framework="",
        features=None,
        config=None,
    ):
        self.framework = framework
        self.config = config if config else {}
        self.features = self.process_features(features)
        self.config = filter_config(self.config, self.name, self.DEFAULTS, self.OPTIONAL, self.REQUIRED)
        self.artifacts = []
        self.supported_fmts = []
        self.tuner = None

    def __repr__(self):
        name = type(self).name
        return f"Backend({name})"

    def process_features(self, features):
        # Filter out non-backend features
        if features is None:
            return []
        features = get_matching_features(features, FeatureType.BACKEND)
        for feature in features:
            assert feature.name in self.FEATURES, f"Incompatible feature: {feature.name}"
            feature.used = True
            feature.add_backend_config(self.name, self.config)
        return features

    @abstractmethod
    def load_model(self, model, input_shapes=None, output_shapes=None, input_types=None, output_types=None):
        pass

    @abstractmethod
    def generate(self) -> Tuple[dict, dict]:
        return {}, {}

    def generate_artifacts(self) -> List[Artifact]:
        start_time = time.time()
        artifacts, metrics = self.generate()
        # TODO: do something with out?
        end_time = time.time()
        diff = end_time - start_time
        if len(metrics) == 0:
            metrics = {"default": Metrics()}
        for name, metrics_ in metrics.items():
            if name == "default":
                metrics_.add("Build Stage Time [s]", diff, True)
            content = metrics_.to_csv(include_optional=True)  # TODO: store df instead?
            artifact = Artifact("build_metrics.csv", content=content, fmt=ArtifactFormat.TEXT, flags=["metrics"])
            if name not in artifacts:
                artifacts[name] = []
            artifacts[name].append(artifact)
        self.artifacts = artifacts
        return artifacts

    @property
    def has_tuner(self):
        return self.tuner is not None

    @property
    def needs_target(self):
        return False

    def set_tuning_records(self, filepath):
        if not self.has_tuner:
            raise NotImplementedError("Backend does not support autotuning")

    def export_artifacts(self, path):
        assert len(self.artifacts) > 0, "No artifacts found, please run generate_artifacts() first"

        if not isinstance(path, Path):
            path = Path(path)

        is_dir = len(path.suffix) == 0
        if is_dir:
            assert (
                path.is_dir()
            ), "The supplied path does not exists."  # Make sure it actually exists (we do not create it by default)
            for artifact in self.artifacts:
                extract = artifact.fmt == ArtifactFormat.MLF
                artifact.export(path, extract=extract)
                # TODO: move the following to a helper function and share code
                # dest = path / artifact.name
                # with open(dest, "w") as outfile:
                #     logger.info(f"Exporting artifact: {artifact.name}")
                #     outfile.write(artifact.content)
        else:
            assert path.parent.is_dir(), "The parent directory does not exist. Make sure to create if beforehand."
            # Warning: the first artifact is considered as main
            # We need to ensure that all further artifacts share the same prefix
            main_prefix = None
            for artifact in self.artifacts:
                if not main:
                    main_prefix = artifact.name
                else:
                    if main_prefix not in artifact.name:
                        logger.warn(
                            f"Skipping export of artifact '{artifact.name}' to prevent overwriting random files"
                        )
                        continue
                dest = path / artifact.name
                with open(dest, "w") as outfile:
                    logger.info(f"Exporting artifact: {artifact.name}")
                    outfile.write(artifact.content)

    def get_platform_config(self, platform):
        return {}

    def add_platform_config(self, platform, config):
        config.update(self.get_platform_config(platform))

    def get_platform_defs(self, platform):
        if platform == "espidf":
            backend_upper = self.name.upper()
            return {f"MLONMCU_BACKEND_{backend_upper}": True}
        else:
            return {"MLONMCU_BACKEND": self.name}

    def add_platform_defs(self, platform, defs):
        defs.update(self.get_platform_defs(platform))

    def supports_model(self, model):
        return any(fmt in self.supported_formats for fmt in model.formats)


def get_parser(backend_name, features, required, defaults):
    # TODO: add help strings should start with a lower case letter
    parser = argparse.ArgumentParser(
        description=f"Run {backend_name} backend",
        formatter_class=argparse.RawTextHelpFormatter,
        # epilog="""Use environment variables to overwrite default paths:""",
    )
    parser.add_argument("model", metavar="MODEL", type=str, nargs=1, help="Model to process")
    parser.add_argument(
        "--output",
        "-o",
        metavar="DIR",
        type=str,
        default=os.path.join(os.getcwd(), "out"),  # TODO: keep this or require flag?
        help="""Output directory/file (default: %(default)s)""",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        help="Print detailed messages for easier debugging (default: %(default)s)",
    )
    parser.add_argument(
        "--print",
        "-p",
        action="store_true",
        help="Print the generated code to the command line instead (default: %(default)s)",  # TODO: instead or both?
    )
    parser.add_argument(
        "-f",
        "--feature",
        type=str,
        metavar="FEATURE",
        # nargs=1,
        action="append",
        choices=list(dict.fromkeys(features)),
        help="Enabled features for the backend (default: %(default)s choices: %(choices)s)",
    )
    parser.add_argument(
        "-c",
        "--config",
        metavar="KEY=VALUE",
        nargs="+",
        action="append",
        help="""Set a number of key-value pairs

Allowed options:
"""
        + "\n".join(
            [f"- [{backend_name}].{key} (Default: {value})" for key, value in defaults.items()]
            + [f"- {key} (required)" for key in required]
        ),
    )
    return parser


def init_backend_features(names, config):
    raise NotImplementedError
    # features = []
    # for name in names:
    #     feature_classes = get_supported_features(feature_type=FeatureType.BACKEND, feature_name=name)
    #     for feature_class in feature_classes:
    #         features.append(feature_class(config=config))
    # return features


def main(backend, args=None):
    parser = get_parser(
        backend.name,
        features=backend.FEATURES,
        required=backend.REQUIRED,
        defaults=backend.DEFAULTS,
    )
    if args:
        args = parser.parse_args(args)
    else:
        args = parser.parse_args()
    # TODO: handle args!
    model = Path(args.model[0])
    config = extract_config(args)
    features_names = extract_feature_names(args)
    features = init_backend_features(features_names, config)
    backend_inst = backend(features=features, config=config)
    backend_inst.load_model(model)
    if args.verbose:
        config["print_outputs"] = True
    backend_inst.generate_artifacts()
    if args.print:
        print("Printing generated artifacts:")
        for artifact in backend_inst.artifacts:
            print(f"=== {artifact.name} ===")
            artifact.print_summary()
            print("=== End ===")
            print()
    else:
        out = args.output
        backend_inst.export_code(out)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of MLonMCU features and the feature registry."""

import re
import pandas as pd
from typing import Union
from pathlib import Path

from mlonmcu.utils import is_power_of_two, filter_none
from mlonmcu.config import str2bool, str2list
from mlonmcu.artifact import Artifact, ArtifactFormat
from .feature import (
    BackendFeature,
    FrameworkFeature,
    PlatformFeature,
    FrontendFeature,
    TargetFeature,
    SetupFeature,
    RunFeature,
    FeatureBase,
)

# from mlonmcu.flow import SUPPORTED_TVM_BACKENDS
SUPPORTED_TVM_BACKENDS = [
    "tvmaot",
    "tvmaotplus",
    "tvmrt",
    "tvmcg",
    "tvmllvm",
]  # Workaround for cirvular import until we have a backend registry


REGISTERED_FEATURES = {}
FEATURE_DEPS = {}


def register_feature(name, depends=None):
    """Decorator for adding a feature to the global registry."""
    if depends is None:
        depends = []

    def real_decorator(obj):
        REGISTERED_FEATURES[name] = obj
        FEATURE_DEPS[name] = depends

    return real_decorator


def get_available_feature_names(feature_type=None):
    """Utility for getting feature names."""
    ret = []
    if feature_type is None:
        return REGISTERED_FEATURES.keys()

    for name, feature in REGISTERED_FEATURES.items():
        if feature_type in list(feature.types()):
            ret.append(name)
    return ret


def get_available_features(feature_type=None, feature_name=None, deps=False):
    """Utility for looking up features."""
    names = get_available_feature_names(feature_type=feature_type)
    names = [name for name in names if feature_name is None or name == feature_name]
    ret = {}
    if deps:
        for name in names:
            names = list(set(names + FEATURE_DEPS[name]))
    for name in names:
        ret[name] = REGISTERED_FEATURES[name]
    return ret


def get_matching_features(features, feature_type):
    return [feature for feature in features if feature_type in feature.types()]


# @register_feature("debug_arena", depends=["debug"])
@register_feature("debug_arena")
class DebugArena(BackendFeature):
    """Enable verbose printing of arena usage for debugging."""

    def __init__(self, features=None, config=None):
        super().__init__("debug_arena", features=features, config=config)

    def get_backend_config(self, backend):
        assert backend in [
            "tvmaot",
            "tvmaotplus",
            "tvmrt",
            "tflmi",
        ], f"Unsupported feature '{self.name}' for backend '{backend}'"
        return {f"{backend}.debug_arena": self.enabled}


# @register_feature("validate", depends=["debug"])
@register_feature("validate")
class Validate(FrontendFeature, PlatformFeature):
    """Enable validaton of inout and output tensors."""

    DEFAULTS = {
        **FeatureBase.DEFAULTS,
        "allow_missing": True,
        "fail_on_error": None,
    }

    def __init__(self, features=None, config=None):
        super().__init__("validate", features=features, config=config)

    @property
    def allow_missing(self):
        value = self.config["allow_missing"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def fail_on_error(self):
        return self.config["fail_on_error"]

    def get_frontend_config(self, frontend):
        if not self.allow_missing:
            raise NotImplementedError
        return {f"{frontend}.use_inout_data": True}

    def get_platform_config(self, platform):
        assert platform == "mlif", f"Unsupported feature '{self.name}' for platform '{platform}'"
        return filter_none(
            {
                f"{platform}.ignore_data": False,
                f"{platform}.fail_on_error": self.fail_on_error,
            }
        )


@register_feature("muriscvnn")
class Muriscvnn(SetupFeature, FrameworkFeature, PlatformFeature):
    """muRISCV-V NN wrappers for TFLite Micro"""

    DEFAULTS = {
        **FeatureBase.DEFAULTS,
        "use_vext": "AUTO",
        "use_pext": "AUTO",
    }

    REQUIRED = {"muriscvnn.src_dir"}

    def __init__(self, features=None, config=None):
        super().__init__("muriscvnn", features=features, config=config)

    @property
    def muriscvnn_dir(self):
        return str(self.config["muriscvnn.src_dir"])

    @property
    def use_vext(self):
        value = self.config["use_vext"]
        if value == "AUTO" or value is None:
            return value
        if not isinstance(value, (bool, int)):
            value = str2bool(value)
        return "ON" if value else "OFF"

    @property
    def use_pext(self):
        value = self.config["use_pext"]
        if value == "AUTO" or value is None:
            return value
        if not isinstance(value, (bool, int)):
            value = str2bool(value)
        return "ON" if value else "OFF"

    def add_framework_config(self, framework, config):
        assert framework == "tflm", f"Unsupported feature '{self.name}' for framework '{framework}'"
        if f"{framework}.optimized_kernel" in config and config[f"{framework}.optimized_kernel"] not in [
            None,
            "cmsis_nn",
        ]:
            RuntimeError(f"There is already a optimized_kernel selected for framework '{framework}'")
        else:
            config[f"{framework}.optimized_kernel"] = "cmsis_nn"

    def get_platform_defs(self, platform):
        assert platform in ["mlif"], f"Unsupported feature '{self.name}' for platform '{platform}'"
        return {
            "MURISCVNN": self.enabled,
            "MURISCVNN_DIR": self.muriscvnn_dir,
            "MURISCVNN_VEXT": self.use_vext,
            "MURISCVNN_PEXT": self.use_pext,
        }

    def get_required_cache_flags(self):
        ret = {}

        ret["tflmc.exe"] = ["muriscvnn"]
        return ret


@register_feature("cmsisnn")
class Cmsisnn(SetupFeature, FrameworkFeature, PlatformFeature):
    """CMSIS-NN kernels for TFLite Micro"""

    DEFAULTS = {
        **FeatureBase.DEFAULTS,
    }

    REQUIRED = {"cmsisnn.dir", "cmsis.dir"}

    def __init__(self, features=None, config=None):
        super().__init__("cmsisnn", features=features, config=config)

    @property
    def cmsisnn_dir(self):
        return str(self.config["cmsisnn.dir"])

    @property
    def cmsis_dir(self):
        return str(self.config["cmsis.dir"])

    def add_framework_config(self, framework, config):
        assert framework == "tflm", f"Unsupported feature '{self.name}' for framework '{framework}'"
        if f"{framework}.optimized_kernel" in config and config[f"{framework}.optimized_kernel"] not in [
            None,
            "cmsis_nn",
        ]:
            RuntimeError(f"There is already a optimized_kernel selected for framework '{framework}'")
        else:
            config[f"{framework}.optimized_kernel"] = "cmsis_nn"

    def get_platform_defs(self, platform):
        assert platform in ["mlif"], f"Unsupported feature '{self.name}' for platform '{platform}'"
        return {
            "CMSISNN": self.enabled,
            "CMSIS_DIR": self.cmsis_dir,
            "CMSISNN_DIR": self.cmsisnn_dir,
        }

    def get_required_cache_flags(self):
        ret = {}
        ret["tflmc.exe"] = ["cmsisnn"]
        return ret


@register_feature("cmsisnnbyoc")
class CmsisnnByoc(SetupFeature, BackendFeature, PlatformFeature):
    """CMSIS-NN kernels for TVM using BYOC wrappers."""

    DEFAULTS = {
        **FeatureBase.DEFAULTS,
        "mcpu": None,  # mve: cortex-m55, dsp: cortex-m4, cortex-m7, cortex-m33, cortex-m35p
        "mattr": None,  # for +nodsp, +nomve
        "debug_last_error": False,
    }

    REQUIRED = {"cmsisnn.dir", "cmsis.dir"}

    def __init__(self, features=None, config=None):
        super().__init__("cmsisnnbyoc", features=features, config=config)

    @property
    def cmsisnn_dir(self):
        return str(self.config["cmsisnn.dir"])

    @property
    def cmsis_dir(self):
        return str(self.config["cmsis.dir"])

    @property
    def mcpu(self):
        return self.config["mcpu"]

    @property
    def mattr(self):
        return self.config["mattr"]

    @property
    def debug_last_error(self):
        return str2bool(self.config["debug_last_error"])

    def add_backend_config(self, backend, config):
        assert backend in SUPPORTED_TVM_BACKENDS, f"Unsupported feature '{self.name}' for backend '{backend}'"
        extras = config.get(f"{backend}.extra_targets", [])
        if extras is None:
            extras = []
        if "cmsis-nn" not in extras:
            if isinstance(extras, str):
                extras = str2list(extras)
            extras.append("cmsis-nn")
        config[f"{backend}.extra_targets"] = extras
        # Ideally cmsisnnbyoc would have a mvei/dsp feature which could be used to set this automatically
        extra_target_details = config.get(f"{backend}.extra_target_details", {})
        if extra_target_details is None:
            extra_target_details = {}
        cmsisnn_target_details = extra_target_details.get("cmsis-nn", {})
        if self.mcpu:
            cmsisnn_target_details["mcpu"] = self.mcpu
        if self.mattr:
            cmsisnn_target_details["mattr"] = self.mattr
        if self.debug_last_error is not None:
            cmsisnn_target_details["debug_last_error"] = self.debug_last_error
        extra_target_details["cmsis-nn"] = cmsisnn_target_details

        config[f"{backend}.extra_target_details"] = extra_target_details

    def get_platform_defs(self, platform):
        assert platform in ["mlif"], f"Unsupported feature '{self.name}' for platform '{platform}'"
        return {
            "CMSISNN": self.enabled,
            "CMSIS_DIR": self.cmsis_dir,
            "CMSISNN_DIR": self.cmsisnn_dir,
        }

    def get_required_cache_flags(self):
        ret = {}
        ret["tvm.build_dir"] = ["cmsisnn"]
        return ret


@register_feature("muriscvnnbyoc")
class MuriscvnnByoc(SetupFeature, BackendFeature, PlatformFeature):
    """MuRiscvNN kernels for TVM using BYOC wrappers."""

    DEFAULTS = {
        **FeatureBase.DEFAULTS,
        "mcpu": None,  # mve: cortex-m55, dsp: cortex-m4, cortex-m7, cortex-m33, cortex-m35p
        "mattr": None,  # for +nodsp, +nomve
        "debug_last_error": False,
        "use_vext": "AUTO",
        "use_pext": "AUTO",
    }

    REQUIRED = {"muriscvnn.src_dir"}

    def __init__(self, features=None, config=None):
        super().__init__("muriscvnnbyoc", features=features, config=config)

    @property
    def muriscvnn_dir(self):
        return str(self.config["muriscvnn.src_dir"])

    @property
    def mcpu(self):
        return self.config["mcpu"]

    @property
    def mattr(self):
        return self.config["mattr"]

    @property
    def debug_last_error(self):
        return str2bool(self.config["debug_last_error"])

    @property
    def use_vext(self):
        value = self.config["use_vext"]
        if value == "AUTO" or value is None:
            return value
        if not isinstance(value, (bool, int)):
            value = str2bool(value)
        return "ON" if value else "OFF"

    @property
    def use_pext(self):
        value = self.config["use_pext"]
        if value == "AUTO" or value is None:
            return value
        if not isinstance(value, (bool, int)):
            value = str2bool(value)
        return "ON" if value else "OFF"

    def add_backend_config(self, backend, config):
        assert backend in SUPPORTED_TVM_BACKENDS, f"Unsupported feature '{self.name}' for backend '{backend}'"
        extras = config.get(f"{backend}.extra_targets", [])
        if extras is None:
            extras = []
        if "cmsis-nn" not in extras:
            if isinstance(extras, str):
                extras = str2list(extras)
            extras.append("cmsis-nn")
        config[f"{backend}.extra_targets"] = extras
        # Ideally cmsisnnbyoc would have a mvei/dsp feature which could be used to set this automatically
        extra_target_details = config.get(f"{backend}.extra_target_details", {})
        if extra_target_details is None:
            extra_target_details = {}
        cmsisnn_target_details = extra_target_details.get("cmsis-nn", {})
        if self.mcpu:
            cmsisnn_target_details["mcpu"] = self.mcpu
        if self.mattr:
            cmsisnn_target_details["mattr"] = self.mattr
        if self.debug_last_error is not None:
            cmsisnn_target_details["debug_last_error"] = self.debug_last_error
        extra_target_details["cmsis-nn"] = cmsisnn_target_details

        config[f"{backend}.extra_target_details"] = extra_target_details

    def get_platform_defs(self, platform):
        assert platform in ["mlif"], f"Unsupported feature '{self.name}' for platform '{platform}'"
        return {
            "MURISCVNN": self.enabled,
            "MURISCVNN_DIR": self.muriscvnn_dir,
            "MURISCVNN_VEXT": self.use_vext,
            "MURISCVNN_PEXT": self.use_pext,
        }

    def get_required_cache_flags(self):
        ret = {}
        ret["tvm.build_dir"] = ["cmsisnn"]
        return ret


VEXT_MIN_ALLOWED_VLEN = 64


# @before_feature("muriscvnn")  # TODO: implement something like this
@register_feature("vext")
class Vext(SetupFeature, TargetFeature, PlatformFeature):
    """Enable vector extension for supported RISC-V targets"""

    DEFAULTS = {
        **FeatureBase.DEFAULTS,
        # None -> use target setting
        "vlen": None,  # 64 does not work with every toolchain
        "elen": None,  # some toolchains may generate auto-vectorized programs with elen 64
        # use target-side settings by default
        "spec": None,
        "embedded": None,
    }

    def __init__(self, features=None, config=None):
        super().__init__("vext", features=features, config=config)

    @property
    def vlen(self):
        value = self.config["vlen"]
        return None if value is None else int(value)

    @property
    def elen(self):
        value = self.config["elen"]
        return None if value is None else int(value)

    @property
    def spec(self):
        return self.config["spec"]

    @property
    def embedded(self):
        return self.config["embedded"]

    def get_target_config(self, target):
        # TODO: enforce llvm toolchain using add_compile_config and CompileFeature?
        assert is_power_of_two(self.vlen)
        assert self.vlen >= VEXT_MIN_ALLOWED_VLEN
        return filter_none(
            {
                f"{target}.enable_vext": True,
                f"{target}.vlen": self.vlen,
                f"{target}.elen": self.elen,
                f"{target}.vext_spec": self.spec,
                f"{target}.embedded_vext": self.embedded,
            }
        )

    # It would be great if we could enforce an llvm toolchain here
    # def add_compile_config(self, config):
    #     # TODO: enforce llvm toolchain using add_compile_config and CompileFeature?
    #     if "mlif.toolchain" in config:
    #         assert "mlif.toolchain" == "llvm", "Vext requires LLVM target sw"
    #     else:
    #         config["mlif.toolchain"] = "llvm"

    def get_platform_defs(self, platform):
        assert platform in ["mlif"], f"Unsupported feature '{self.name}' for platform '{platform}'"
        return {
            "RISCV_VEXT": self.enabled,
            "RISCV_VLEN": self.vlen,
        }

    def get_required_cache_flags(self):
        return {
            "muriscvnn.lib": ["vext"],
            "tflmc.exe": ["vext"],
            # "riscv_gcc.install_dir": ["vext"],
            # "riscv_gcc.name": ["vext"],
        }


@register_feature("pext")
class Pext(SetupFeature, TargetFeature, PlatformFeature):
    """Enable packed SIMD extension for supported RISC-V targets"""

    DEFAULTS = {
        **FeatureBase.DEFAULTS,
        # use target-side settings by default
        "spec": None,
    }

    def __init__(self, features=None, config=None):
        super().__init__("pext", features=features, config=config)

    @property
    def spec(self):
        return self.config["spec"]

    def get_target_config(self, target):
        return filter_none(
            {
                f"{target}.enable_pext": True,  # Handle via arch characters in the future
                f"{target}.pext_spec": self.spec,
            }
        )

    def get_platform_defs(self, platform):
        assert platform in ["mlif"], f"Unsupported feature '{self.name}' for platform '{platform}'"
        return {"RISCV_PEXT": self.enabled}

    def get_required_cache_flags(self):
        # These will be merged automatically with existing ones
        return {
            "muriscvnn.lib": ["pext"],
            "tflmc.exe": ["pext"],
            "riscv_gcc.install_dir": ["pext"],
            "riscv_gcc.name": ["pext"],
        }


@register_feature("bext")
class Bext(SetupFeature, TargetFeature, PlatformFeature):
    """Enable bitmanipulation extension for supported RISC-V targets"""

    DEFAULTS = {
        **FeatureBase.DEFAULTS,
        # use target-side settings by default
        "spec": None,
        "zba": True,
        "zbb": True,
        "zbc": True,
        "zbs": True,
    }

    def __init__(self, features=None, config=None):
        super().__init__("bext", features=features, config=config)

    @property
    def spec(self):
        return self.config["spec"]

    @property
    def zba(self):
        value = self.config["zba"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def zbb(self):
        value = self.config["zbb"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def zbc(self):
        value = self.config["zbc"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def zbs(self):
        value = self.config["zbs"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    def get_target_config(self, target):
        return filter_none(
            {
                f"{target}.enable_bext": True,  # Handle via arch characters in the future
                f"{target}.bext_spec": self.spec,
                f"{target}.bext_zba": self.zba,
                f"{target}.bext_zbb": self.zbb,
                f"{target}.bext_zbc": self.zbc,
                f"{target}.bext_zbs": self.zbs,
            }
        )

    def get_platform_defs(self, platform):
        assert platform in ["mlif"], f"Unsupported feature '{self.name}' for platform '{platform}'"
        return {"RISCV_PEXT": self.enabled}

    def get_required_cache_flags(self):
        # These will be merged automatically with existing ones
        return {
            # "riscv_gcc.install_dir": ["bext"],
            # "riscv_gcc.name": ["bext"],
            "riscv_gcc.install_dir": [],
            "riscv_gcc.name": [],
        }


@register_feature("debug")
class Debug(SetupFeature, PlatformFeature):
    """Enable debugging ability of target software."""

    def __init__(self, features=None, config=None):
        super().__init__("debug", features=features, config=config)

    def get_required_cache_flags(self):
        return {} if self.enabled else {}  # TODO: remove?

    def get_platform_config(self, platform):
        return {f"{platform}.debug": self.enabled}


@register_feature("gdbserver")
class GdbServer(TargetFeature):
    """Start debugging session for target software using gdbserver."""

    DEFAULTS = {
        **FeatureBase.DEFAULTS,
        "attach": None,
        "port": None,
    }

    def __init__(self, features=None, config=None):
        super().__init__("gdbserver", features=features, config=config)

    @property
    def attach(self):
        value = self.config["attach"]
        return str2bool(value, allow_none=True) if not isinstance(value, (bool, int)) else value

    @property
    def port(self):
        return int(self.config["port"]) if self.config["port"] is not None else None

    def get_target_config(self, target):
        assert target in ["host_x86", "etiss_pulpino", "etiss", "ovpsim", "corev_ovpsim"]
        return filter_none(
            {
                f"{target}.gdbserver_enable": self.enabled,
                f"{target}.gdbserver_attach": self.attach,
                f"{target}.gdbserver_port": self.port,
            }
        )


@register_feature("etissdbg")
class ETISSDebug(SetupFeature, TargetFeature):
    """Debug ETISS internals."""

    def __init__(self, features=None, config=None):
        super().__init__("etissdbg", features=features, config=config)

    def get_required_cache_flags(self):
        return {"etiss.install_dir": ["dbg"], "etissvp.script": ["dbg"]} if self.enabled else {}

    def get_target_config(self, target):
        assert target in ["etiss_pulpino", "etiss"]
        return {f"{target}.debug_etiss": self.enabled}


@register_feature("trace")
class Trace(TargetFeature):
    """Enable tracing of all memory accesses in ETISS."""

    DEFAULTS = {**FeatureBase.DEFAULTS, "to_file": True}  # ETISS can only trace to file

    @property
    def to_file(self):
        value = self.config["to_file"]
        return str2bool(value, allow_none=True) if not isinstance(value, (bool, int)) else value

    def __init__(self, features=None, config=None):
        super().__init__("trace", features=features, config=config)

    # def add_target_config(self, target, config, directory=None):
    def add_target_config(self, target, config):
        assert target in ["etiss_pulpino", "etiss", "ovpsim"]
        if target in ["etiss_pulpino", "etiss"]:
            config.update({f"{target}.trace_memory": self.enabled})
        elif target == "ovpsim":
            extra_args_new = config.get("extra_args", [])
            extra_args_new.append("--trace --tracemem SAX")
            if self.to_file:
                # assert directory is not None
                directory = Path(".")  # Need to use relative path because target.dir not available here
                trace_file = directory / "trace.txt"
                extra_args_new.append("--tracefile")
                extra_args_new.append(trace_file)
            config.update({f"{target}.extra_args": extra_args_new})


@register_feature("unpacked_api")
class UnpackedApi(BackendFeature):  # TODO: should this be a feature or config only?
    """Use unpacked interface api for TVMAOT backend to reduce stack usage."""

    def __init__(self, features=None, config=None):
        super().__init__("unpacked_api", features=features, config=config)

    def get_backend_config(self, backend):
        assert backend in ["tvmaot"], f"Unsupported feature '{self.name}' for backend '{backend}'"
        return {f"{backend}.unpacked_api": self.enabled}


@register_feature("packed")
class Packed(FrameworkFeature, FrontendFeature, BackendFeature, SetupFeature):
    """Sub-8-bit and sparsity feature for TFLite Micro kernels."""

    def __init__(self, features=None, config=None):
        super().__init__("packed", features=features, config=config)

    def get_framework_config(self, framework):
        raise NotImplementedError

    def get_frontend_config(self, frontend):
        assert frontend in ["tflm"], f"Unsupported feature '{self.name} for frontend '{frontend}''"
        return {f"{frontend}.use_packed_weights": self.enabled}

    def get_backend_config(self, backend):
        raise NotImplementedError

    def get_required_cache_flags(self):
        return {"tflmc.exe": ["packed"]}


@register_feature("packing")
class Packing(FrontendFeature):
    """Sub-8-bit and sparse weight packing for TFLite Frontend."""

    def __init__(self, features=None, config=None):
        super().__init__("packing", features=features, config=config)

    def get_frontend_config(self, frontend):
        assert frontend in ["tflm"], f"Unsupported feature '{self.name} for frontend '{frontend}''"
        raise NotImplementedError
        return {f"{frontend}.pack_weights": self.enabled}


@register_feature("usmp")
class Usmp(BackendFeature):
    """Unified Static Memory Planning algorithm integrated in TVM"""

    DEFAULTS = {
        **FeatureBase.DEFAULTS,
        "algorithm": "greedy_by_conflicts",  # options: greedy_by_conflicts, greedy_by_size, hill_climb
    }

    def __init__(self, features=None, config=None):
        super().__init__("usmp", features=features, config=config)

    @property
    def algorithm(self):
        return str(self.config["algorithm"])

    def add_backend_config(self, backend, config):
        assert backend in ["tvmaot"], f"Unsupported feature '{self.name}' for backend '{backend}'"
        if f"{backend}.extra_pass_config" in config:
            tmp = config[f"{backend}.extra_pass_config"]
        elif "extra_pass_config" in config:
            tmp = config["extra_pass_config"]
        else:
            tmp = {}
        if isinstance(tmp, str):
            import ast

            tmp = ast.literal_eval(tmp)
        assert isinstance(tmp, dict)
        tmp["tir.usmp.enable"] = self.enabled
        if self.algorithm in ["greedy_by_size", "greedy_by_conflicts", "hill_climb"]:
            tmp["tir.usmp.algorithm"] = self.algorithm
        else:
            tmp["tir.usmp.custom_algorithm"] = self.algorithm
        config.update(
            {f"{backend}.extra_pass_config": tmp, f"{backend}.arena_size": 0}
        )  # In recent TVM versions USMP will have it's own arena.

    # -> enable this via backend


@register_feature("fuse_ops")
class FuseOps(BackendFeature):
    """TODO"""

    DEFAULTS = {
        **FeatureBase.DEFAULTS,
        "max_depth": 100,
    }

    def __init__(self, features=None, config=None):
        super().__init__("fuse_ops", features=features, config=config)

    @property
    def max_depth(self):
        return int(self.config["max_depth"])

    def add_backend_config(self, backend, config):
        # assert backend in ["tvmaot"], f"Unsupported feature '{self.name}' for backend '{backend}'"
        # TODO: tvm only
        if f"{backend}.extra_pass_config" in config:
            tmp = config[f"{backend}.extra_pass_config"]
        elif "extra_pass_config" in config:
            tmp = config["extra_pass_config"]
        else:
            tmp = {}
        if isinstance(tmp, str):
            import ast

            tmp = ast.literal_eval(tmp)
        assert isinstance(tmp, dict)
        tmp["relay.FuseOps.max_depth"] = self.max_depth
        config.update({f"{backend}.extra_pass_config": tmp})


@register_feature("moiopt")
class MOIOPT(BackendFeature):
    """Memory-Optimizing, Inter-Operator Tiling - currently only supported with custom TVM"""

    DEFAULTS = {
        **FeatureBase.DEFAULTS,
        "noftp": False,
        "onlyftp": False,
        "norecurse": False,
        "maxpartitions": 0,
    }

    def __init__(self, features=None, config=None):
        super().__init__("moiopt", features=features, config=config)

    def add_backend_config(self, backend, config):
        assert backend in ["tvmaot", "tvmrt"], f"Unsupported feature '{self.name}' for backend '{backend}'"
        if f"{backend}.extra_pass_config" in config:
            tmp = config[f"{backend}.extra_pass_config"]
        elif "extra_pass_config" in config:
            tmp = config["extra_pass_config"]
        else:
            tmp = {}
        tmp["relay.moiopt.enable"] = self.enabled
        tmp["relay.moiopt.noftp"] = self.config["noftp"]
        tmp["relay.moiopt.onlyftp"] = self.config["onlyftp"]
        tmp["relay.moiopt.norecurse"] = self.config["norecurse"]
        tmp["relay.moiopt.maxpartitions"] = self.config["maxpartitions"]
        config.update({f"{backend}.extra_pass_config": tmp})

    # -> enable this via backend


@register_feature("visualize")
class Visualize(FrontendFeature):
    """Visualize TFLite models."""

    DEFAULTS = {
        **FeatureBase.DEFAULTS,
    }

    REQUIRED = {"tflite_visualize.exe"}

    def __init__(self, features=None, config=None):
        super().__init__("visualize", features=features, config=config)

    @property
    def tflite_visualize_exe(self):
        return self.config["tflite_visualize.exe"]

    def get_frontend_config(self, frontend):
        assert frontend in ["tflite"], f"Unsupported feature '{self.name}' for frontend '{frontend}'"
        return filter_none(
            {
                f"{frontend}.visualize_enable": self.enabled,
                f"{frontend}.visualize_script": self.tflite_visualize_exe,
            }
        )

    def update_formats(self, frontend, input_formats, output_formats):
        assert frontend in ["tflite"], f"Unsupported feature '{self.name}' for frontend '{frontend}'"
        if self.enabled:
            output_formats.append(ArtifactFormat.TEXT)


@register_feature("relayviz")
class Relayviz(FrontendFeature):
    """Visualize TVM relay models."""

    DEFAULTS = {
        **FeatureBase.DEFAULTS,
        "plotter": "term",  # Alternative: dot
    }

    def __init__(self, features=None, config=None):
        super().__init__("relayviz", features=features, config=config)

    @property
    def plotter(self):
        return self.config.get("plotter", None)

    def get_frontend_config(self, frontend):
        assert frontend in ["relay"], f"Unsupported feature '{self.name}' for frontend '{frontend}'"
        return filter_none(
            {
                f"{frontend}.visualize_graph": self.enabled,
                f"{frontend}.relayviz_plotter": self.plotter,
            }
        )

    def update_formats(self, frontend, input_formats, output_formats):
        assert frontend in ["relay"], f"Unsupported feature '{self.name}' for frontend '{frontend}'"
        if self.enabled:
            output_formats.append(ArtifactFormat.TEXT)


@register_feature("autotuned")
class Autotuned(BackendFeature):
    """Use existing TVM autotuning logs in backend."""

    # TODO: FronendFeature to collect tuning logs or will we store them somewhere else?

    DEFAULTS = {
        **FeatureBase.DEFAULTS,
        "mode": "autotvm",  # further options: autoscheduler, metascheduler
        "results_file": None,
    }

    def __init__(self, features=None, config=None):
        super().__init__("autotuned", features=features, config=config)

    @property
    def results_file(self):
        return self.config.get("results_file", None)

    @property
    def mode(self):
        value = self.config["mode"]
        assert value in ["autotvm", "autoscheduler", "metascheduler"]
        return value

    def get_backend_config(self, backend):
        assert backend in SUPPORTED_TVM_BACKENDS
        # TODO: error handling her eor on backend?
        return filter_none(
            {
                f"{backend}.use_tuning_results": self.enabled,
                f"{backend}.autotuned_results_file": self.results_file,
                f"{backend}.autotuned_mode": self.mode,
            }
        )


@register_feature("autotune")
class Autotune(RunFeature):
    """Generic autotuning feature for enabling the TUNE stage only."""

    def __init__(self, features=None, config=None):
        super().__init__("autotune", features=features, config=config)

    def get_run_config(self):
        return {"run.tune_enabled": self.enabled}


# not registered!
class TVMTuneBase(PlatformFeature):
    DEFAULTS = {
        **FeatureBase.DEFAULTS,
        "results_file": None,
        "append": None,
        "trials": None,
        "trials_single": None,
        "early_stopping": None,
        "num_workers": None,
        "max_parallel": None,
        "use_rpc": None,
        "timeout": None,
        "visualize": None,
        "visualize_file": None,
        "visualize_live": None,
        "tasks": None,
        # All None to use the defaults defined in the backend instead
    }

    @property
    def results_file(self):
        return self.config["results_file"] if "results_file" in self.config else None

    @property
    def append(self):
        return self.config["append"] if "append" in self.config else None

    @property
    def trials(self):
        return self.config["trials"] if "trials" in self.config else None

    @property
    def trials_single(self):
        return self.config["trials_single"] if "trials_single" in self.config else None

    @property
    def early_stopping(self):
        return self.config["early_stopping"] if "early_stopping" in self.config else None

    @property
    def num_workers(self):
        return self.config["num_workers"] if "num_workers" in self.config else None

    @property
    def max_parallel(self):
        return self.config["max_parallel"] if "max_parallel" in self.config else None

    @property
    def use_rpc(self):
        return self.config["use_rpc"] if "use_rpc" in self.config else None

    @property
    def timeout(self):
        return self.config["timeout"] if "timeout" in self.config else None

    @property
    def visualize(self):
        return self.config["visualize"]

    @property
    def visualize_file(self):
        return self.config["visualize_file"]

    @property
    def visualize_live(self):
        return self.config["visualize_live"]

    @property
    def tasks(self):
        return self.config["tasks"]

    def get_platform_config(self, platform):
        assert platform in ["tvm", "microtvm"]
        # TODO: figure out a default path automatically
        return filter_none(
            {
                f"{platform}.autotuning_results_file": self.results_file,
                f"{platform}.autotuning_append": self.append,
                f"{platform}.autotuning_trials": self.trials,
                f"{platform}.autotuning_trials_single": self.trials_single,
                f"{platform}.autotuning_early_stopping": self.early_stopping,
                f"{platform}.autotuning_num_workers": self.num_workers,
                f"{platform}.autotuning_max_parallel": self.max_parallel,
                f"{platform}.autotuning_timeout": self.timeout,
                f"{platform}.autotuning_visualize": self.visualize,
                f"{platform}.autotuning_visualize_file": self.visualize_file,
                f"{platform}.autotuning_visualize_live": self.visualize_live,
                f"{platform}.autotuning_tasks": self.tasks,
            }
        )


@register_feature("autotvm", depends=["autotune"])
class AutoTVM(TVMTuneBase):
    """Use the TVM autotuner inside the backend to generate tuning logs."""

    # TODO: graphtuner
    # TODO: tuner base feature class

    DEFAULTS = {
        **TVMTuneBase.DEFAULTS,
        "tuner": None,
    }

    def __init__(self, features=None, config=None):
        super().__init__("autotvm", features=features, config=config)

    @property
    def tuner(self):
        return self.config["tuner"] if "tuner" in self.config else None

    def get_platform_config(self, platform):
        ret = super().get_platform_config(platform)
        new = filter_none(
            {
                f"{platform}.autotvm_enable": self.enabled,
                f"{platform}.autotvm_tuner": self.tuner,
            }
        )
        ret.update(new)
        return ret


@register_feature("autoscheduler", depends=["autotune"])
class AutoScheduler(TVMTuneBase):
    """TODO"""

    # TODO: metascheduler
    # TODO: graphtuner
    # TODO: tuner base feature class

    DEFAULTS = {
        **TVMTuneBase.DEFAULTS,
        "include_simple_tasks": None,
        "log_estimated_latency": None,
    }

    def __init__(self, features=None, config=None):
        super().__init__("autoschedule", features=features, config=config)

    @property
    def include_simple_tasks(self):
        return self.config["include_simple_tasks"]

    @property
    def log_estimated_latency(self):
        return self.config["log_estimated_latency"]

    def get_platform_config(self, platform):
        ret = super().get_platform_config(platform)
        new = filter_none(
            {
                f"{platform}.autoscheduler_enable": self.enabled,
                f"{platform}.autoscheduler_include_simple_tasks": self.include_simple_tasks,
                f"{platform}.autoscheduler_log_estimated_latency": self.log_estimated_latency,
            }
        )
        ret.update(new)
        return ret


@register_feature("metascheduler", depends=["autotune"])
class MetaScheduler(TVMTuneBase):
    """TODO"""

    DEFAULTS = {
        **TVMTuneBase.DEFAULTS,
    }

    def __init__(self, features=None, config=None):
        super().__init__("metascheduler", features=features, config=config)

    def get_platform_config(self, platform):
        ret = super().get_platform_config(platform)
        new = filter_none(
            {
                f"{platform}.metascheduler_enable": self.enabled,
            }
        )
        ret.update(new)
        return ret


@register_feature("disable_legalize")
class DisableLegalize(BackendFeature, SetupFeature):
    """Enable transformation to reduces sizes of intermediate buffers by skipping legalization passes."""

    REQUIRED = {"tvm_extensions.wrapper"}

    def __init__(self, features=None, config=None):
        super().__init__("disable_legalize", features=features, config=config)

    @property
    def tvm_extensions_wrapper(self):
        return self.config["tvm_extensions.wrapper"]

    def add_backend_config(self, backend, config):
        assert backend in SUPPORTED_TVM_BACKENDS, f"Unsupported feature '{self.name}' for backend '{backend}'"
        if f"{backend}.tvmc_extra_args" in config:
            config[f"{backend}.tvmc_extra_args"].append("--disable-legalize")
        else:
            config[f"{backend}.tvmc_extra_args"] = ["--disable-legalize"]
        if f"{backend}.tvmc_custom_script" in config:
            assert config[f"{backend}.tvmc_custom_script"] is None or str(
                config[f"{backend}.tvmc_custom_script"]
            ) == str(
                self.tvm_extensions_wrapper
            ), f"{backend}.tvmc_custom_script is already set. Can't enable feature: {self.name}"
        config[f"{backend}.tvmc_custom_script"] = self.tvm_extensions_wrapper

    def get_required_cache_flags(self):
        ret = {}

        ret["tvm.pythonpath"] = ["patch"]
        return ret


@register_feature("uma_backends")
class UMABackends(BackendFeature):
    """Add directories that contain UMA backends."""

    REQUIRED = set()

    DEFAULTS = {
        **FeatureBase.DEFAULTS,
        "uma_dir": "",
        "uma_target": "",
    }

    def __init__(self, features=None, config=None):
        super().__init__("uma_backends", features=features, config=config)

    @property
    def uma_dir(self):
        return self.config["uma_dir"]

    @property
    def uma_target(self):
        return self.config["uma_target"]

    def add_backend_config(self, backend, config):
        assert backend in SUPPORTED_TVM_BACKENDS, f"Unsupported feature '{self.name}' for backend '{backend}'"
        tvmcArgs = ["--experimental-tvmc-extension", self.uma_dir]
        if f"{backend}.tvmc_extra_args" in config:
            config[f"{backend}.tvmc_extra_args"].extend(tvmcArgs)
        else:
            config[f"{backend}.tvmc_extra_args"] = tvmcArgs
        extras = config.get(f"{backend}.extra_target", [])
        if self.uma_target not in extras:
            if isinstance(extras, str):
                extras = [extras]
            extras.append(self.uma_target)
        config[f"{backend}.extra_target"] = extras


@register_feature("demo")
class Demo(PlatformFeature):
    """Run demo application instead of benchmarking code."""

    DEFAULTS = {
        **FeatureBase.DEFAULTS,
        "board": None,
        "print_stats": False,
        "print_interval_ms": 5000,
    }

    def __init__(self, features=None, config=None):
        super().__init__("demo", features=features, config=config)

    @property
    def board(self):
        return self.config["board"]

    @property
    def print_stats(self):
        return self.config["print_stats"]

    @property
    def print_interval_ms(self):
        return self.config["print_interval_ms"]

    def get_platform_defs(self, platform):
        assert platform in ["espidf"], f"Unsupported feature '{self.name}' for platform '{platform}'"
        # TODO: espidf.demo_mode, disable wdt, runtime stats,
        return {}

    def get_platform_config(self, platform):
        assert platform in ["espidf"], f"Unsupported feature '{self.name}' for platform '{platform}'"
        # TODO: espidf.demo_mode, disable wdt, runtime stats,
        return {}


@register_feature("cachesim")
class CacheSim(TargetFeature):
    """Collect information on cache misses etc. with spike target"""

    DEFAULTS = {
        **FeatureBase.DEFAULTS,
        "ic_enable": False,
        "ic_config": "64:8:32",
        "dc_enable": False,
        "dc_config": "64:8:32",
        "l2_enable": False,
        "l2_config": "262144:8:32",  # TODO: find a meaningful value
        "log_misses": False,
        "detailed": False,
    }

    def __init__(self, features=None, config=None):
        super().__init__("cachesim", features=features, config=config)

    @property
    def ic_enable(self):
        value = self.config["ic_enable"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def ic_config(self):
        return self.config["ic_config"]

    @property
    def dc_enable(self):
        value = self.config["dc_enable"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def dc_config(self):
        return self.config["dc_config"]

    @property
    def l2_enable(self):
        value = self.config["l2_enable"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def l2_config(self):
        return self.config["l2_config"]

    @property
    def log_misses(self):
        value = self.config["log_misses"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def detailed(self):
        value = self.config["detailed"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    # def add_target_config(self, target, config, directory=None):
    def add_target_config(self, target, config):
        assert target in ["spike"], f"Unsupported feature '{self.name}' for target '{target}'"
        if self.enabled:
            spike_args = config.get(f"{target}.extra_args", [])
            if self.ic_enable:
                assert self.ic_config is not None and len(self.ic_config) > 0
                spike_args.append(f"--ic={self.ic_config}")
            if self.dc_enable:
                assert self.dc_config is not None and len(self.dc_config) > 0
                spike_args.append(f"--dc={self.dc_config}")
            if self.l2_enable:
                assert self.l2_config is not None and len(self.l2_config) > 0
                spike_args.append(f"--l2={self.l2_config}")
            if self.log_misses:
                spike_args.append("--log-cache-miss")
            config.update({f"{target}.extra_args": spike_args})

    def get_target_callbacks(self, target):
        assert target in ["spike"], f"Unsupported feature '{self.name}' for target '{target}'"
        if self.enabled:

            def cachesim_callback(stdout, metrics, artifacts, directory=None):
                """Callback which parses the targets output and updates the generated metrics and artifacts."""
                expr = (
                    r"(D|I|L2)\$ ((?:Bytes (?:Read|Written))|(?:Read|Write) "
                    r"(?:Accesses|Misses)|(?:Writebacks)|(?:Miss Rate)):\s*(\d+\.?\d*%?)*"
                )
                matches = re.compile(expr).findall(stdout)
                prefixes = [
                    x for (x, y) in zip(["I", "D", "L2"], [self.ic_enable, self.dc_enable, self.l2_enable]) if y
                ]
                for groups in matches:
                    assert len(groups) == 3
                    prefix, label, value = groups
                    if not self.detailed:
                        if "Rate" not in label:
                            continue
                    value = int(value) if "%" not in value else float(value[:-1]) / 100
                    if prefix in prefixes:
                        for m in metrics:
                            m.add(f"{prefix}-Cache {label}", value)
                return stdout

            return None, cachesim_callback


@register_feature("log_instrs")
class LogInstructions(TargetFeature):
    """Enable logging of the executed instructions of a simulator-based target."""

    DEFAULTS = {**FeatureBase.DEFAULTS, "to_file": False}

    def __init__(self, features=None, config=None):
        super().__init__("log_instrs", features=features, config=config)

    @property
    def to_file(self):
        value = self.config["to_file"]
        return str2bool(value, allow_none=True) if not isinstance(value, (bool, int)) else value

    # def add_target_config(self, target, config, directory=None):
    def add_target_config(self, target, config):
        assert target in ["spike", "etiss_pulpino", "etiss", "ovpsim", "corev_ovpsim", "gvsoc_pulp"]
        if not self.enabled:
            return
        if target == "spike":
            extra_args_new = config.get("extra_args", [])
            extra_args_new.append("-l")
            if self.to_file:
                # assert directory is not None
                directory = Path(".")  # Need to use relative path because target.dir not available here
                log_file = directory / "instrs.txt"
                extra_args_new.append(f"--log={log_file}")
            config.update({f"{target}.extra_args": extra_args_new})
        elif target in ["etiss_pulpino", "etiss"]:
            plugins_new = config.get("plugins", [])
            plugins_new.append("PrintInstruction")
            config.update({f"{target}.plugins": plugins_new})
        elif target in ["ovpsim", "corev_ovpsim"]:
            extra_args_new = config.get("extra_args", [])
            extra_args_new.append("--trace")
            if self.to_file:
                # assert directory is not None
                directory = Path(".")  # Need to use relative path because target.dir not available here
                log_file = directory / "instrs.txt"
                extra_args_new.append("--tracefile")
                extra_args_new.append(log_file)
            config.update({f"{target}.extra_args": extra_args_new})
        elif target == "gvsoc_pulp":
            extra_args_new = config.get("extra_args", [])
            if self.to_file:
                # assert directory is not None
                directory = Path(".")  # Need to use relative path because target.dir not available here
                log_file = directory / "instrs.txt"
                extra_args_new.append(f"--trace=insn:{log_file}")
            else:
                extra_args_new.append("--trace=insn")
            config.update({f"{target}.extra_args": extra_args_new})

    def get_target_callbacks(self, target):
        assert target in [
            "spike",
            "etiss_pulpino",
            "etiss",
            "ovpsim",
            "corev_ovpsim",
            "gvsoc_pulp",
        ], f"Unsupported feature '{self.name}' for target '{target}'"
        if self.enabled:
            if not target == "gvsoc_pulp":

                def log_instrs_callback(stdout, metrics, artifacts, directory=None):
                    """Callback which parses the targets output and updates the generated metrics and artifacts."""
                    new_lines = []
                    if self.to_file:
                        if target in ["etiss_pulpino", "etiss"]:
                            # TODO: update stdout and remove log_instrs lines
                            instrs = []
                            for line in stdout.split("\n"):
                                if target in ["etiss_pulpino", "etiss"]:
                                    expr = re.compile(r"0x[a-fA-F0-9]+: .* \[.*\]")
                                match = expr.match(line)
                                if match is not None:
                                    instrs.append(line)
                                else:
                                    new_lines.append(line)
                            content = "\n".join(instrs)
                            stdout = "\n".join(new_lines)
                        else:
                            assert target in ["spike", "ovpsim", "corev_ovpsim"]
                            log_file = Path(directory) / "instrs.txt"
                            with open(log_file, "r") as f:
                                content = f.read()
                        instrs_artifact = Artifact(
                            f"{target}_instrs.log",
                            content=content,
                            fmt=ArtifactFormat.TEXT,
                            flags=(self.name, target),
                        )
                        artifacts.append(instrs_artifact)
                    return stdout

                return None, log_instrs_callback
        return None, None


@register_feature("arm_mvei", depends=["arm_dsp"])
class ArmMvei(SetupFeature, TargetFeature, PlatformFeature):
    """Enable MVEI extension for supported ARM targets"""

    def __init__(self, features=None, config=None):
        super().__init__("arm_mvei", features=features, config=config)

    def get_target_config(self, target):
        assert target in ["corstone300"]
        return {
            f"{target}.enable_mvei": True,  # TODO: remove if not required (only enforce m33/m55)
        }

    def get_required_cache_flags(self):
        return {
            "cmsisnn.lib": ["mvei"],
            "tflmc.exe": ["mvei"],
        }

    def get_platform_defs(self, platform):
        return {"ARM_MVEI": self.enabled}


@register_feature("arm_dsp")
class ArmDsp(SetupFeature, TargetFeature, PlatformFeature):
    """Enable DSP extension for supported ARM targets"""

    def __init__(self, features=None, config=None):
        super().__init__("arm_dsp", features=features, config=config)

    def get_target_config(self, target):
        assert target in ["corstone300"]
        return {
            f"{target}.enable_dsp": True,  # TODO: remove if not required (only enforce m33/m55)
        }

    def get_required_cache_flags(self):
        # These will be merged automatically with existing ones
        return {
            "cmsisnn.lib": ["dsp"],
            "tflmc.exe": ["dsp"],
        }

    def get_platform_defs(self, platform):
        return {"ARM_DSP": self.enabled}


@register_feature("target_optimized")
class TargetOptimized(RunFeature):
    """Overwrite backend options according to chosen target."""

    DEFAULTS = {
        **FeatureBase.DEFAULTS,
        "layouts": True,
        "schedules": True,
    }

    @property
    def layouts(self):
        value = self.config["layouts"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def schedules(self):
        value = self.config["schedules"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    def __init__(self, features=None, config=None):
        super().__init__("target_optimized", features=features, config=config)

    def get_run_config(self):
        if self.enabled:
            return {
                "run.target_optimized_layouts": self.layouts,
                "run.target_optimized_schedules": self.schedules,
            }
        else:
            return {}


# Needs: vext
# RISC-V only
# Warning: Auto-vectorization is turned on by default quite low optimization levels
# Therfore this feature is mainly for debugging the auto-vectorization procedure
@register_feature("auto_vectorize")
class AutoVectorize(PlatformFeature):
    """Enable auto_vectorization for supported MLIF platform targets."""

    DEFAULTS = {
        **FeatureBase.DEFAULTS,
        "verbose": False,
        "loop": True,
        "slp": True,
        "force_vector_width": None,  # llvm only
        "force_vector_interleave": None,  # llvm only
        "custom_unroll": False,  # TODO: this is not related to vectorization -> move to llvm toolchain!
    }

    def __init__(self, features=None, config=None):
        super().__init__("auto_vectorize", features=features, config=config)

    @property
    def verbose(self):
        value = self.config["verbose"]
        # return str2bool(value) if not isinstance(value, (bool, int)) else value
        if value is None or not value:
            return "OFF"
        assert isinstance(value, str)
        value = value.lower()
        assert value in ["loop", "slp", "none"]
        return value

    @property
    def loop(self):
        value = self.config["loop"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def slp(self):
        value = self.config["slp"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def force_vector_width(self):
        value = self.config["force_vector_width"]
        # bool not allowed!
        if value is None:
            return "OFF"
        if isinstance(value, str):
            value = int(value)
        assert isinstance(value, int)
        if value <= 1:
            return "OFF"
        return value

    @property
    def force_vector_interleave(self):
        value = self.config["force_vector_interleave"]
        # bool not allowed!
        if value is None:
            return "OFF"
        if isinstance(value, str):
            value = int(value)
        assert isinstance(value, int)
        if value <= 1:
            return "OFF"
        return value

    @property
    def custom_unroll(self):
        value = self.config["custom_unroll"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    def get_platform_defs(self, platform):
        return {
            "RISCV_AUTO_VECTORIZE": self.enabled,
            "RISCV_AUTO_VECTORIZE_VERBOSE": self.verbose,
            "RISCV_AUTO_VECTORIZE_LOOP": self.loop and self.enabled,
            "RISCV_AUTO_VECTORIZE_SLP": self.slp and self.enabled,
            "RISCV_AUTO_VECTORIZE_FORCE_VECTOR_WIDTH": self.force_vector_width,
            "RISCV_AUTO_VECTORIZE_FORCE_VECTOR_INTERLEAVE": self.force_vector_interleave,
            "RISCV_AUTO_VECTORIZE_CUSTOM_UNROLL": self.custom_unroll,
        }


@register_feature("benchmark")
class Benchmark(PlatformFeature, TargetFeature):
    """Profile code using supported platforms."""

    # TODO: would make sense to move end_to_end here as well!

    DEFAULTS = {
        **FeatureBase.DEFAULTS,
        "num_runs": 1,
        "num_repeat": 1,
        "total": False,
        "aggregate": "avg",  # Allowed: avg, max, min, none, all
    }

    def __init__(self, features=None, config=None):
        super().__init__("benchmark", features=features, config=config)

    @property
    def num_runs(self):
        return int(self.config["num_runs"])

    @property
    def num_repeat(self):
        return int(self.config["num_repeat"])

    @property
    def total(self):
        value = self.config["total"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def aggregate(self):
        value = self.config["aggregate"]
        assert value in ["avg", "all", "max", "min", "none"]
        return value

    def get_platform_config(self, platform):
        supported = ["mlif", "tvm", "microtvm"]  # TODO: support espidf
        assert platform in supported, f"Unsupported feature '{self.name}' for platform '{platform}'"

        if platform in ["tvm", "microtvm"]:
            return {
                f"{platform}.number": self.num_runs,
                f"{platform}.repeat": self.num_repeat,
                f"{platform}.aggregate": self.aggregate,
                f"{platform}.total_time": self.total,
            }
        else:
            return {}

    def get_target_config(self, target):
        return {
            f"{target}.repeat": self.num_repeat,
        }

    def get_platform_defs(self, platform):
        supported = ["mlif", "espidf", "tvm", "microtvm", "zephyr"]  # TODO: support microtvm and espidf
        assert platform in supported, f"Unsupported feature '{self.name}' for platform '{platform}'"

        if platform == "mlif":
            return {"NUM_RUNS": self.num_runs}
        elif platform == "espidf":
            return {"MLONMCU_NUM_RUNS": self.num_runs}
        else:
            return {}

    def get_target_callbacks(self, target):
        if self.enabled:

            def benchmark_callback(stdout, metrics, artifacts, directory=None):
                if len(metrics) <= 1:
                    return stdout
                metrics_ = metrics[1:]  # drop first run (warmup)

                # TODO: this currently processes all numeric metrics, should probably ignore stuff like MIPS etc.
                data_ = [
                    {
                        key: (float(value) / self.num_runs) if self.num_runs > 1 else value
                        for key, value in m.data.items()
                        if "cycle" in key.lower() or "time" in key.lower()
                    }
                    for m in metrics_
                ]

                df = pd.DataFrame(data_)

                if self.aggregate == "all":
                    aggs = ["mean", "min", "max"]
                elif self.aggregate in ["avg", "mean"]:
                    aggs = ["mean"]
                elif self.aggregate == "min":
                    aggs = ["min"]
                elif self.aggregate == "max":
                    aggs = ["max"]
                elif self.aggregate == "none":
                    aggs = []

                if len(aggs) == 0:
                    data = {}
                else:
                    df_ = df.agg(aggs)

                    # rename columns
                    index_mapping = {
                        "mean": "Average",
                        "min": "Min",
                        "max": "Max",
                    }

                    df_ = df_.rename(index=index_mapping)

                    data = df_.to_dict()

                    data = {f"{prefix} {key}": value for key, temp in data.items() for prefix, value in temp.items()}
                if self.total:
                    data.update(
                        {
                            f"Total {key}": (value * self.num_runs) if self.num_runs > 1 else value
                            for key, value in data_[-1].items()
                            if "cycle" in key.lower() or "time" in key.lower()
                        }
                    )
                metrics_ = metrics_[-1]
                metrics_.data.update(data)

                for key in data_[-1].keys():
                    if key in metrics_.order:
                        if self.total:
                            metrics_.order.append(f"Total {key}")
                        metrics_.order.remove(key)
                for key in data.keys():
                    if key not in metrics_.order:
                        metrics_.order.append(key)
                metrics.clear()
                metrics.append(metrics_)
                return stdout

            benchmark_callback.priority = 0

            return None, benchmark_callback


@register_feature("tvm_rpc")
class TvmRpc(PlatformFeature):
    """Run TVM models on a RPC device."""

    DEFAULTS = {**FeatureBase.DEFAULTS, "hostname": None, "port": None, "key": None}  # tracker

    def __init__(self, features=None, config=None):
        super().__init__("tvm_rpc", features=features, config=config)

    @property
    def use_rpc(self):
        return self.config["use_rpc"]

    @property
    def hostname(self):
        return self.config["hostname"]

    @property
    def port(self):
        return self.config["port"]

    @property
    def key(self):
        return self.config["key"]

    def get_platform_config(self, platform):
        assert platform in ["tvm", "microtvm"]
        return filter_none(
            {
                f"{platform}.use_rpc": self.enabled,
                f"{platform}.rpc_hostname": self.hostname,
                f"{platform}.rpc_port": self.port,
                f"{platform}.rpc_key": self.key,
            }
        )


@register_feature("tvm_profile")
class TvmProfile(PlatformFeature):
    """Profile code using TVM Platform."""

    def __init__(self, features=None, config=None):
        super().__init__("tvm_profile", features=features, config=config)

    def get_platform_config(self, platform):
        supported = ["tvm", "microtvm"]
        assert platform in supported, f"Unsupported feature '{self.name}' for platform '{platform}'"
        return {
            f"{platform}.profile": self.enabled,
        }


@register_feature("xcorev")
class XCoreV(TargetFeature, PlatformFeature, SetupFeature):
    DEFAULTS = {
        **FeatureBase.DEFAULTS,
        "mac": True,
        "mem": True,
        "bi": True,
        "alu": True,
        "bitmanip": True,
        "simd": True,
        "hwlp": True,
    }

    def __init__(self, features=None, config=None):
        super().__init__("xcorev", features=features, config=config)

    @property
    def mac(self):
        value = self.config["mac"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def mem(self):
        value = self.config["mem"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def bi(self):
        value = self.config["bi"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def alu(self):
        value = self.config["alu"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def bitmanip(self):
        value = self.config["bitmanip"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def simd(self):
        value = self.config["simd"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def hwlp(self):
        value = self.config["hwlp"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    # def add_target_config(self, target, config, directory=None):
    def add_target_config(self, target, config):
        assert target in [
            "etiss",
            "microtvm_etiss",
            "corev_ovpsim",
            "cv32e40p",
        ], f"Unsupported feature '{self.name}' for target '{target}'"
        if self.enabled:
            config[f"{target}.enable_xcorevmac"] = self.mac
            config[f"{target}.enable_xcorevmem"] = self.mem
            config[f"{target}.enable_xcorevbi"] = self.bi
            config[f"{target}.enable_xcorevalu"] = self.alu
            config[f"{target}.enable_xcorevbitmanip"] = self.bitmanip
            config[f"{target}.enable_xcorevsimd"] = self.simd
            config[f"{target}.enable_xcorevhwlp"] = self.hwlp


@register_feature("xpulp")
class Xpulp(TargetFeature, PlatformFeature, SetupFeature):
    DEFAULTS = {
        **FeatureBase.DEFAULTS,
        "xpulp_version": 2,
        "nopostmod": False,
        "noindregreg": False,
        "novect": False,
        "nohwloop": False,
        "hwloopmin": 2,
        "hwloopalign": False,
        "nomac": False,
        "nopartmac": False,
        "nominmax": False,
        "noabs": False,
        "nobitop": False,
        "nosext": False,
        "noclip": False,
        "noaddsubnormround": False,
        "noshufflepack": False,
        "nomulmacnormround": False,
        "noshufflepack": False,
    }

    REQUIRED = {"pulp_gcc.install_dir", "pulp_gcc.name"}

    def __init__(self, features=None, config=None):
        super().__init__("xpulp", features=features, config=config)

    # Except the "enabled" in FeatureBase.DEFAULTS and "xpulp_version"
    # every key in DEFAULTS should in principle have a getter function with the same name as the key
    # These getter functions will be stored in getter_functions array.
    # Default getter function for the keys whose corresponding value has bool type:
    # def getter_bool(self, key_name):
    #    return self.generalized_str2bool(self.config[key_name])
    # Default getter function for the keys whose corresponding value has int type:
    # def getter_int(self, key_name):
    #    return int(self.config["<key_name>"])
    # No default, i.e. customized getter functions are defined separately

    # Default getter functions:
    @staticmethod
    def generalized_str2bool(input: Union[str, bool, int]) -> bool:
        return str2bool(input) if isinstance(input, str) else input

    def getter_bool(self, key_name: Union[str, bool, int]) -> bool:
        return self.generalized_str2bool(self.config[key_name])

    def getter_int(self, key_name: Union[str, bool, int]) -> int:
        return int(self.config[key_name])

    # No default, i.e. customized getter functions are defined in the following (now empty)

    # custom_config_getter contains customized @property function which do not follow the pattern above.
    getter_functions = {
        "nopostmod": getter_bool,
        "noindregreg": getter_bool,
        "novect": getter_bool,
        "nohwloop": getter_bool,
        "hwloopmin": getter_int,
        "hwloopalign": getter_bool,
        "nomac": getter_bool,
        "nopartmac": getter_bool,
        "nominmax": getter_bool,
        "noabs": getter_bool,
        "nobitop": getter_bool,
        "nosext": getter_bool,
        "noclip": getter_bool,
        "noaddsubnormround": getter_bool,
        "noshufflepack": getter_bool,
        "nomulmacnormround": getter_bool,
        "noshufflepack": getter_bool,
    }

    @property
    def xpulp_version(self):
        value = self.config["xpulp_version"]
        value = int(value) if not isinstance(value, int) else value  # convert to int
        assert value in [None, 2, 3], f"xpulp_version must be None, 2 or 3, but get {value}"
        return value

    def get_platform_defs(self, platform):
        # The following create EXTRA_FLAGS (type is str) for gcc
        # example
        # {"nopostmod": True, "novect": True, ...} ==> EXTRA_FLAGS = "-mnopostmod -mnovect ..."
        EXTRA_FLAGS = ""
        for key in self.getter_functions:
            if isinstance(self.getter_functions[key](self, key), bool):
                if self.getter_functions[key](self, key):
                    EXTRA_FLAGS += f" -m{key}"
                continue
            if isinstance(self.getter_functions[key](self, key), int):
                EXTRA_FLAGS += f" -m{key}={self.getter_functions[key](self, key)}"
                continue
        EXTRA_FLAGS = "'" + EXTRA_FLAGS.strip() + "'"
        return {
            "EXTRA_C_FLAGS": EXTRA_FLAGS,
            "EXTRA_CXX_FLAGS": EXTRA_FLAGS,
            "EXTRA_ASM_FLAGS": EXTRA_FLAGS,
        }

    def add_platform_defs(self, platform, defs):
        addition_defs = self.get_platform_defs(platform)
        self.merge_dicts(defs, addition_defs)

    @staticmethod
    def merge_dicts(dict1, dict2):
        """
        This function tries to merge dict1 and dict2 into dict1
        :param dict1: A dictionary
        :param dict2: A dictionary to be added
        :return: Void
        Example 1:
        dict1 = {"a": 1, "b": "hello", "c": [1, 2, 3]}
        dict2 = {"f": 3, "b": "world", "c": [4, 5, 6]}
        merge_dicts(dict1, dict2)
        print(dict1)
        ==>
        {"a": 1, "b": "hello world", "c": [1, 2, 3, 4, 5, 6], "f":3}
        Note: Here "hello" and "world" are merged as two string join.
        Here [1,2,3] and [4,5,6] are merged as list addition
        Example 2:
        dict1 = {"a": 1}
        dict2 = {"a": 3}
        merge_dicts(dict1, dict2)
        ==>
        RuntimeError: The method to merge a: 1 and a: 3 is not defined
        """
        for key in dict2.keys():
            if key in dict1.keys():
                dict1_value = dict1[key]
                dict2_value = dict2[key]
                if isinstance(dict1_value, (str, list)) and type(dict1_value) is type(dict2_value):
                    if isinstance(dict1_value, str):
                        dict1[key] = dict1_value + " " + dict2_value
                    else:
                        dict1[key] = dict1_value + dict2_value
                else:
                    raise RuntimeError(
                        f"The method to merge {key}: {dict1_value} and {key}: {dict2_value} is not defined"
                    )
            else:
                dict1[key] = dict2[key]

    def get_target_config(self, target):
        return filter_none({f"{target}.xpulp_version": self.xpulp_version})


@register_feature("split_layers")
class SplitLayers(FrontendFeature):
    """Split TFLite models into subruns."""

    REQUIRED = {"tflite_pack.exe"}

    def __init__(self, features=None, config=None):
        super().__init__("split_layers", features=features, config=config)

    @property
    def tflite_pack_exe(self):
        return self.config["tflite_pack.exe"]

    def get_frontend_config(self, frontend):
        assert frontend in ["tflite"], f"Unsupported feature '{self.name}' for frontend '{frontend}'"
        return filter_none(
            {
                f"{frontend}.split_layers": self.enabled,
                f"{frontend}.pack_script": self.tflite_pack_exe,
            }
        )


@register_feature("tflite_analyze")
class TfLiteAnalyze(FrontendFeature):
    """Get the estimated ROM, RAM and MACs from a TFLite model."""

    REQUIRED = {"tflite_analyze.exe"}

    def __init__(self, features=None, config=None):
        super().__init__("tflite_analyze", features=features, config=config)

    @property
    def tflite_analyze_exe(self):
        return self.config["tflite_analyze.exe"]

    def get_frontend_config(self, frontend):
        assert frontend in ["tflite"], f"Unsupported feature '{self.name}' for frontend '{frontend}'"
        return filter_none(
            {
                f"{frontend}.analyze_enable": self.enabled,
                f"{frontend}.analyze_script": self.tflite_analyze_exe,
            }
        )

    def update_formats(self, frontend, input_formats, output_formats):
        assert frontend in ["tflite"], f"Unsupported feature '{self.name}' for frontend '{frontend}'"
        if self.enabled:
            output_formats.append(ArtifactFormat.TEXT)


# @register_feature("hpmcounter")
class HpmCounter(TargetFeature, PlatformFeature):  # TODO: SetupFeature?
    """Use RISC-V Performance Counters"""

    DEFAULTS = {
        **FeatureBase.DEFAULTS,
        "num_counters": 32,
        "supported_counters": 1,  # To check if number of enabled counters exceeds counters implemented in hw
        "enabled_counters": [],
        "counter_names": [],
    }

    # def __init__(self, features=None, config=None):
    #     super().__init__("hpmcounter", features=features, config=config)

    @property
    def num_counters(self):
        temp = self.config["num_counters"]
        return int(temp)

    @property
    def supported_counters(self):
        temp = self.config["supported_counters"]
        return int(temp)

    @property
    def enabled_counters(self):
        temp = self.config["enabled_counters"]
        if isinstance(temp, int):
            temp = [temp]
        elif isinstance(temp, str):
            temp = str2list(temp)
        assert isinstance(temp, list)
        temp = list(map(int, temp))
        return temp

    @property
    def counter_names(self):
        temp = self.config["counter_names"]
        if not isinstance(temp, list):
            temp = str2list(temp)
        return temp

    def get_platform_defs(self, platform):
        assert platform in ["mlif"], f"Unsupported feature '{self.name}' for platform '{platform}'"
        assert self.supported_counters >= len(self.enabled_counters)
        assert max(self.enabled_counter) < self.supported_counters
        return {
            "HPM_COUNTERS": self.num_counters,
            **({f"USE_HPM{i}": i in self.enabled_counters for i in range(self.num_counters)}),
        }

    def get_target_callbacks(self, target):
        if self.enabled:

            def hpm_callback(stdout, metrics, artifacts, directory=None):
                """Callback for extracting HPM metrics from stdout"""
                print("stdout", stdout)
                # TODO: add metrics
                # TODO: remove HPM lines from stdout
                return stdout

            return None, hpm_callback
        return None, None


@register_feature("cv32_hpmcounter")
class CV32HpmCounter(HpmCounter):  # TODO: SetupFeature?
    """Use RISC-V Performance Counters"""

    DEFAULTS = {
        **HpmCounter.DEFAULTS,
        "num_counters": 12,
        "supported_counters": 32,  # TODO
        "enabled_counters": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
        "counter_names": [
            "Cycles",
            "Instructions",
            "LD Stalls",
            "JMP Stalls",
            "IMiss",
            "LD",
            "ST",
            "Jump",
            "Branch",
            "Branch Taken",
            "Compressed",
            "Pipe Stall",
        ],
    }

    def __init__(self, features=None, config=None):
        super().__init__("cv32_hpmcounter", features=features, config=config)


@register_feature("vanilla_accelerator")
class VanillaAccelerator(TargetFeature):
    """TODO"""

    DEFAULTS = {
        **FeatureBase.DEFAULTS,
        "plugin_name": "VanillaAccelerator",
        "base_addr": None,
    }

    def __init__(self, features=None, config=None):
        super().__init__("vanilla_accelerator", features=features, config=config)

    @property
    def plugin_name(self):
        value = self.config["plugin_name"]
        return value

    @property
    def base_addr(self):
        value = self.config["base_addr"]
        return value

    def add_target_config(self, target, config):
        assert target in ["etiss"]
        if not self.enabled:
            return
        plugins_new = config.get(f"{target}.plugins", [])
        plugins_new.append(self.plugin_name)
        config.update({f"{target}.plugins": plugins_new})
        if self.base_addr is not None:
            extra_plugin_config = config.get(f"{target}.extra_plugin_config", {})
            assert self.name not in extra_plugin_config
            extra_plugin_config[self.name]["baseaddr"] = self.base_addr
            config.update({f"{target}.extra_plugin_config": extra_plugin_config})



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import os
import tempfile
from pathlib import Path
from typing import Tuple

from .backend import TFLMBackend
import mlonmcu.setup.utils as utils
from mlonmcu.config import str2bool
from mlonmcu.flow.backend import main
from mlonmcu.logging import get_logger
from mlonmcu.artifact import Artifact, ArtifactFormat

logger = get_logger()


class TFLMCBackend(TFLMBackend):
    name = "tflmc"

    FEATURES = {"debug_arena"}

    DEFAULTS = {
        **TFLMBackend.DEFAULTS,
        "print_outputs": False,
        "custom_ops": [],
        "registrations": {},
        "debug_arena": False,
    }

    REQUIRED = TFLMBackend.REQUIRED | {"tflmc.exe"}

    @property
    def print_outputs(self):
        value = self.config["print_outputs"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    def __init__(self, features=None, config=None):
        super().__init__(features=features, config=config)
        self.model_data = None
        self.prefix = "model"  # Without the _
        self.artifacts = (
            []
        )  # TODO: either make sure that ony one model is processed at a time or move the artifacts to the methods
        # TODO: decide if artifacts should be handled by code (str) or file path or binary data

    def generate_header(self):
        upper_prefix = self.prefix.upper()
        code = f"""
// This file is generated. Do not edit.
#ifndef {upper_prefix}_GEN_H
#define {upper_prefix}_GEN_H

#include <stddef.h>

#ifdef __cplusplus
extern "C" {{
#endif

void model_init();
void *model_input_ptr(int index);
size_t model_input_size(int index);
size_t model_inputs();
void model_invoke();
void *model_output_ptr(int index);
size_t model_output_size(int index);
size_t model_outputs();

#ifdef __cplusplus
}}
#endif

#endif  // {upper_prefix}_GEN_H
"""
        return code

    def generate(self) -> Tuple[dict, dict]:
        artifacts = []
        assert self.model is not None
        tflmc_exe = None
        if "tflmc.exe" in self.config:
            tflmc_exe = self.config["tflmc.exe"]
        else:
            # Lookup cache
            raise NotImplementedError
        with tempfile.TemporaryDirectory() as tmpdirname:
            logger.debug("Using temporary directory for codegen results: %s", tmpdirname)
            args = []
            args.append(str(self.model))
            args.append(str(Path(tmpdirname) / f"{self.prefix}.cc"))
            args.append(f"{self.prefix}_")
            out = utils.exec_getout(tflmc_exe, *args, live=self.print_outputs)
            files = [f for f in os.listdir(tmpdirname) if os.path.isfile(os.path.join(tmpdirname, f))]
            # TODO: ensure that main file is processed first
            for filename in files:
                with open(Path(tmpdirname) / filename, "r") as handle:
                    content = handle.read()
                    artifacts.append(Artifact(filename, content=content, fmt=ArtifactFormat.SOURCE))
            header_content = self.generate_header()
            header_artifact = Artifact(f"{self.prefix}.cc.h", content=header_content, fmt=ArtifactFormat.SOURCE)
            artifacts.append(header_artifact)
            stdout_artifact = Artifact("tflmc_out.log", content=out, fmt=ArtifactFormat.TEXT)
            artifacts.append(stdout_artifact)

        return {"default": artifacts}, {}


if __name__ == "__main__":
    sys.exit(
        main(
            TFLMCBackend,
            args=sys.argv[1:],
        )
    )  # pragma: no cover



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from pathlib import Path

from mlonmcu.utils import filter_none
from mlonmcu.config import str2bool
from mlonmcu.target.target import Target
from mlonmcu.target.riscv.util import sort_extensions_canonical, join_extensions
from mlonmcu.feature.features import SUPPORTED_TVM_BACKENDS

from mlonmcu.logging import get_logger

from .microtvm_template_target import TemplateMicroTvmPlatformTarget

logger = get_logger()


class MlonmcuMicroTvmPlatformTarget(TemplateMicroTvmPlatformTarget):
    FEATURES = Target.FEATURES | {"xcorev"}  # TODO: for now just etiss, future: any?

    DEFAULTS = {
        **Target.DEFAULTS,
        # "project_type": "?",
        "verbose": False,
        "quiet": False,
        "workspace_size_bytes": None,
        "toolchain": "gcc",
        "xlen": 32,
        "extensions": ["i", "m", "a", "c"],  # TODO overwrite extensions elegantly
        "fpu": "double",  # allowed: none, single, double
        "arch": None,
        "abi": None,
        "attr": "",
        "cpu_arch": None,
        "etiss_extra_args": "",
        "enable_xcorevmac": False,
        "enable_xcorevmem": False,
        "enable_xcorevbi": False,
        "enable_xcorevalu": False,
        "enable_xcorevbitmanip": False,
        "enable_xcorevsimd": False,
        "enable_xcorevhwlp": False,
    }
    REQUIRED = Target.REQUIRED | {
        "mlif.src_dir",
        "riscv_gcc.install_dir",
        "riscv_gcc.name",
        "etissvp.script",
        "llvm.install_dir",
    }

    def __init__(self, name=None, features=None, config=None):
        super().__init__(name=name, features=features, config=config)
        self.template_path = self.microtvm_etiss_src_dir / "template_project"
        self.option_names = [
            # "extra_files_tar",
            # "project_type",
            "verbose",
            "quiet",
            "workspace_size_bytes",
            # "warning_as_error",
            # "compile_definitions",
            # "config_main_stack_size",
            # "etissvp_script_args",
            # "transport",
        ]

    @property
    def microtvm_etiss_src_dir(self):
        return Path(self.config["microtvm_etiss.src_dir"])

    @property
    def riscv_gcc_install_dir(self):
        return Path(self.config["riscv_gcc.install_dir"])

    @property
    def riscv_gcc_name(self):
        return self.config["riscv_gcc.name"]

    @property
    def etiss_script(self):
        return Path(self.config["etissvp.script"])

    @property
    def etiss_extra_args(self):
        return self.config["etiss_extra_args"]

    @property
    def enable_xcorevmac(self):
        value = self.config["enable_xcorevmac"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def enable_xcorevmem(self):
        value = self.config["enable_xcorevmem"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def enable_xcorevbi(self):
        value = self.config["enable_xcorevbi"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def enable_xcorevalu(self):
        value = self.config["enable_xcorevalu"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def enable_xcorevbitmanip(self):
        value = self.config["enable_xcorevbitmanip"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def enable_xcorevsimd(self):
        value = self.config["enable_xcorevsimd"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def enable_xcorevhwlp(self):
        value = self.config["enable_xcorevhwlp"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    def get_project_options(self):
        ret = super().get_project_options()
        # TODO: allow generating etiss ini config instead!
        ret.update(
            {
                "gcc_prefix": self.riscv_gcc_install_dir,
                "gcc_name": self.riscv_gcc_name,
                "llvm_dir": self.llvm_prefix,
                "toolchain": self.toolchain,
                "etiss_script": self.etiss_script,
                "etiss_args": self.etiss_extra_args,
                "arch": self.gcc_arch,
                "abi": self.abi,
                "cpu_arch": self.cpu_arch,
            }
        )
        return ret

    @property
    def xlen(self):
        return int(self.config["xlen"])

    @property
    def fpu(self):
        value = self.config["fpu"]
        if value is None or not value:
            value = "none"
        assert value in ["none", "single", "double"]
        return value

    @property
    def extensions(self):
        exts = self.config.get("extensions", []).copy()
        if not isinstance(exts, list):
            exts = exts.split(",")
        required = []
        required.append("zicsr")
        if "g" not in exts:
            # if self.fpu in ["single", "double"]:
            #     required.append("zicsr")
            if self.fpu == "double":
                required.append("d")
                required.append("f")
                required.append("zifencei")
            if self.fpu == "single":
                required.append("f")
        if "xcorev" not in exts:
            if self.enable_xcorevmac:
                required.append("xcvmac")
            if self.enable_xcorevmem:
                required.append("xcvmem")
            if self.enable_xcorevbi:
                required.append("xcvbi")
            if self.enable_xcorevalu:
                required.append("xcvalu")
            if self.enable_xcorevsimd:
                required.append("xcvsimd")
            if self.enable_xcorevhwlp:
                required.append("xcvhwlp")
        for ext in required:
            if ext not in exts:
                exts.append(ext)
        return exts

    @property
    def llvm_arch(self):
        temp = self.config["arch"]  # TODO: allow underscores and versions
        if temp:
            return temp
        else:
            filtered_exts = [ext for ext in self.extensions if ext not in ["zifencei", "zicsr"]]
            exts_str = join_extensions(sort_extensions_canonical(filtered_exts, lower=True))
            return f"rv{self.xlen}{exts_str}"

    @property
    def gcc_arch(self):
        temp = self.config["arch"]  # TODO: allow underscores and versions
        if temp:
            return temp
        else:
            filtered_exts = [ext for ext in self.extensions if "xcv" not in ext]
            exts_str = join_extensions(sort_extensions_canonical(filtered_exts, lower=True))
            return f"rv{self.xlen}{exts_str}"

    @property
    def abi(self):
        temp = self.config["abi"]
        if temp:
            return temp
        else:
            if self.xlen == 32:
                temp = "ilp32"
            elif self.xlen == 64:
                temp = "lp64"
            else:
                raise RuntimeError(f"Invalid xlen: {self.xlen}")
            if "d" in self.extensions or "g" in self.extensions:
                temp += "d"
            elif "f" in self.extensions:
                temp += "f"
            return temp

    @property
    def attr(self):
        attrs = str(self.config["attr"]).split(",")
        if len(attrs) == 1 and len(attrs[0]) == 0:
            attrs = []
        for ext in sort_extensions_canonical(self.extensions, lower=True, unpack=True):
            if ext in ["i", "zicsr", "zifencei"]:
                continue
            attrs.append(f"+{ext}")
        attrs = list(set(attrs))
        return ",".join(attrs)

    @property
    def cpu_arch(self):
        tmp = self.config.get("cpu_arch", None)
        if tmp is None:
            tmp = f"RV{self.xlen}IMACFD"
        return tmp

    @property
    def llvm_prefix(self):
        return Path(self.config["llvm.install_dir"])

    @property
    def toolchain(self):
        value = self.config["toolchain"]
        assert value in ["gcc", "llvm"]
        return value

    def get_backend_config(self, backend, optimized_layouts=False, optimized_schedules=False):
        ret = {}
        if backend in SUPPORTED_TVM_BACKENDS:
            arch_clean = self.llvm_arch.replace("imafd", "g").replace("_", "-")
            ret.update(
                {
                    "target_march": self.llvm_arch,
                    "target_mtriple": self.riscv_gcc_name,
                    "target_mabi": self.abi,
                    "target_mattr": self.attr,
                    "target_mcpu": f"generic-rv{self.xlen}",
                    "target_model": f"etiss-{arch_clean}",
                }
            )
            if optimized_schedules:
                ret.update(
                    {
                        "target_device": "riscv_cpu",
                        "target_keys": None,
                    }
                )
        return ret

    def add_backend_config(self, backend, config, optimized_layouts=False, optimized_schedules=False):
        new = filter_none(
            self.get_backend_config(
                backend, optimized_layouts=optimized_layouts, optimized_schedules=optimized_schedules
            )
        )

        # only allow overwriting non-none values
        # to support accepting user-vars
        new = {key: value for key, value in new.items() if config.get(key, None) is None}
        config.update(new)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import itertools

from mlonmcu.platform import get_platforms
from mlonmcu.config import resolve_required_config


def get_platform_names(context):
    return context.environment.lookup_platform_configs(names_only=True)


def get_platforms_targets(context, config=None):
    platform_names = get_platform_names(context)
    platform_classes = get_platforms()
    # To initialize the platform we need to provide a config with required paths.
    platforms = []
    config_ = {}
    if context:
        config_.update(context.environment.vars)
    if config:
        config_.update(config)

    for platform_name in platform_names:
        platform_cls = platform_classes[platform_name]
        required_keys = platform_cls.REQUIRED
        new_config = resolve_required_config(
            required_keys,
            config=config_,
            cache=context.cache,
            default_flags=context.environment.flags,
        )
        platform = platform_cls(config=new_config)
        platforms.append(platform)
    return {platform.name: platform.get_supported_targets() for platform in platforms}


def get_platforms_backends(context, config=None):
    platform_names = get_platform_names(context)
    platform_classes = get_platforms()
    # To initialize the platform we need to provide a config with required paths.
    platforms = []
    config_ = {}
    if context:
        config_.update(context.environment.vars)
    if config:
        config_.update(config)

    for platform_name in platform_names:
        platform_cls = platform_classes[platform_name]
        required_keys = platform_cls.REQUIRED
        new_config = resolve_required_config(
            required_keys,
            config=config_,
            cache=context.cache,
            default_flags=context.environment.flags,
        )
        platform = platform_cls(config=new_config)
        platforms.append(platform)
    return {platform.name: platform.get_supported_backends() for platform in platforms}


def print_platforms(platform_names):
    print("Platforms:")
    if len(platform_names) == 0:
        print("No platforms found")
        return
    for name in platform_names:
        print("  - " + name)


def print_targets(platform_targets):
    print("Targets:")
    target_names = set(itertools.chain(*platform_targets.values()))
    if len(target_names) == 0:
        print("No targets found")
        return
    for target_name in target_names:
        target_platforms = [
            platform_name
            for platform_name, platform_targets in platform_targets.items()
            if target_name in platform_targets
        ]
        print("  - " + target_name + " [" + ", ".join(target_platforms) + "]")


def print_backends(platform_backends):
    print("Backends:")
    backend_names = set(itertools.chain(*platform_backends.values()))
    if len(backend_names) == 0:
        print("No targets found")
        return
    for backend_name in backend_names:
        backend_platforms = [
            platform_name
            for platform_name, platform_backends in platform_backends.items()
            if backend_name in platform_backends
        ]
        print("  - " + backend_name + " [" + ", ".join(backend_platforms) + "]")


def print_summary(context):
    print("Platform Targets Summary\n")
    platform_targets = get_platforms_targets(context)
    print_platforms(platform_targets.keys())
    print()
    platform_backends = get_platforms_backends(context)
    print_backends(platform_backends)
    print()
    print_targets(platform_targets)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of tasks used to dynamically install MLonMCU dependencies"""

import os
import multiprocessing

from mlonmcu.setup.task import TaskType
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.setup import utils
from mlonmcu.logging import get_logger

from .common import get_task_factory
from .cv32e40p import _validate_cv32e40p

logger = get_logger()
Tasks = get_task_factory()

########
# ara  #
########


def _validate_verilator(context: MlonMcuContext, params=None):
    return (
        context.environment.has_target("ara")
        or context.environment.has_target("ara_rtl")
        or context.environment.has_target("vicuna")
        or _validate_cv32e40p(context, params=params)
    )


@Tasks.provides(["verilator.src_dir"])
@Tasks.validate(_validate_verilator)
@Tasks.register(category=TaskType.TARGET)
def clone_verilator(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Clone the verilator repository."""
    if not params:
        params = {}
    user_vars = context.environment.vars
    flags = utils.makeFlags()
    verilatorName = utils.makeDirName("verilator", flags=flags)
    verilatorSrcDir = context.environment.paths["deps"].path / "src" / verilatorName
    if "verilator.src_dir" in user_vars:  # TODO: also check command line flags?
        # This would overwrite the cache.ini entry which is NOT wanted! -> return false but populate gcc_name?
        verilatorSrcDir = user_vars["verilator.src_dir"]
    else:
        if rebuild or not utils.is_populated(verilatorSrcDir):
            verilatorRepo = context.environment.repos["verilator"]
            utils.clone_wrapper(verilatorRepo, verilatorSrcDir, refresh=rebuild)
    context.cache["verilator.src_dir", flags] = verilatorSrcDir


@Tasks.needs(["verilator.src_dir"])
@Tasks.provides(["verilator.build_dir", "verilator.install_dir"])
@Tasks.validate(_validate_verilator)
@Tasks.register(category=TaskType.TARGET)
def build_verilator(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Build the verilator simulator."""
    if not params:
        params = {}
    flags = utils.makeFlags()
    verilatorName = utils.makeDirName("verilator", flags=flags)
    verilatorSrcDir = context.environment.paths["deps"].path / "src" / verilatorName
    verilatorBuildDir = context.environment.paths["deps"].path / "build" / verilatorName
    verilatorInstallDir = context.environment.paths["deps"].path / "install" / verilatorName
    utils.mkdirs(verilatorBuildDir)
    user_vars = context.environment.vars
    if "verilator.build_dir" in user_vars:  # TODO: also check command line flags?
        # This would overwrite the cache.ini entry which is NOT wanted! -> return false but populate gcc_name?
        verilatorBuildDir = user_vars["verilator.build_dir"]
    else:
        if rebuild or not utils.is_populated(verilatorBuildDir):
            env = os.environ.copy()
            env["VERILATOR_ROOT"] = ""
            utils.exec_getout(
                "autoconf",
                env=env,
                cwd=verilatorSrcDir,
                live=verbose,
            )
            env = os.environ.copy()
            utils.exec_getout(
                str(verilatorSrcDir / "configure"),
                f"--prefix={verilatorInstallDir}",
                env=env,
                cwd=verilatorBuildDir,
                # live=verbose,
                live=False,
            )
            utils.make(cwd=verilatorBuildDir, threads=threads, live=verbose)
            utils.make("install", cwd=verilatorBuildDir, threads=threads, live=verbose)
    context.cache["verilator.build_dir"] = verilatorBuildDir
    context.cache["verilator.install_dir"] = verilatorInstallDir


@Tasks.needs(["verilator.build_dir"])
@Tasks.provides(["verilator.install_dir"])
@Tasks.validate(_validate_verilator)
@Tasks.register(category=TaskType.TARGET)
def install_verilator(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Install the verilator simulator."""
    if not params:
        params = {}
    flags = utils.makeFlags()
    verilatorName = utils.makeDirName("verilator", flags=flags)
    verilatorBuildDir = context.environment.paths["deps"].path / "build" / verilatorName
    verilatorInstallDir = context.environment.paths["deps"].path / "install" / verilatorName
    utils.mkdirs(verilatorBuildDir)
    user_vars = context.environment.vars
    if "verilator.install_dir" in user_vars:  # TODO: also check command line flags?
        # This would overwrite the cache.ini entry which is NOT wanted! -> return false but populate gcc_name?
        verilatorInstallDir = user_vars["verilator.install_dir"]
    else:
        if rebuild or not utils.is_populated(verilatorInstallDir):
            env = os.environ.copy()
            utils.exec_getout(
                "make",
                "install",
                env=env,
                cwd=verilatorBuildDir,
                live=verbose,
            )
    context.cache["verilator.install_dir"] = verilatorInstallDir
    context.export_paths.add(verilatorInstallDir)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of tasks used to dynamically install MLonMCU dependencies"""

import multiprocessing
from pathlib import Path

from mlonmcu.setup.task import TaskType
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.setup import utils
from mlonmcu.logging import get_logger

from .common import get_task_factory

logger = get_logger()

Tasks = get_task_factory()


def _validate_srecord(context: MlonMcuContext, params={}):
    return context.environment.has_target("cv32e40p") or context.environment.has_target("vicuna")


# TODO: cleanup build dir?


@Tasks.provides(["srecord.src_dir"])
@Tasks.validate(_validate_srecord)
@Tasks.register(category=TaskType.MISC)
def clone_srecord(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Clone the screcord repository."""
    name = utils.makeDirName("srecord")
    user_vars = context.environment.vars
    if "srecord.src_dir" in user_vars:
        srcDir = Path(user_vars["srecord.src_dir"])
        rebuild = False
    else:
        srcDir = context.environment.paths["deps"].path / "src" / name
    if rebuild or not utils.is_populated(srcDir):
        repo = context.environment.repos["srecord"]
        utils.clone_wrapper(repo, srcDir, refresh=rebuild)
    context.cache["srecord.src_dir"] = srcDir


# Only build srec_cat here
# @Tasks.needs(["etiss.src_dir", "llvm.install_dir"])
@Tasks.needs(["srecord.src_dir"])
@Tasks.provides(["srecord.build_dir", "srecord.install_dir"])
@Tasks.validate(_validate_srecord)
@Tasks.register(category=TaskType.MISC)
def build_srecord(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Build the srecord unitilties."""
    name = utils.makeDirName("srecord")
    buildDir = context.environment.paths["deps"].path / "build" / name
    installDir = context.environment.paths["deps"].path / "install" / name
    user_vars = context.environment.vars
    if "srecord.install_dir" in user_vars:
        return False
    utils.mkdirs(installDir)
    if rebuild or not utils.is_populated(buildDir):
        utils.mkdirs(buildDir)
        utils.cmake(
            context.cache["srecord.src_dir"],
            "-DCMAKE_INSTALL_PREFIX=" + str(installDir),
            cwd=buildDir,
            debug=False,
            live=verbose,
        )
        utils.make("srec_cat", cwd=buildDir, threads=threads, live=verbose)
        utils.copy(buildDir / "srec_cat" / "srec_cat", installDir / "srec_cat")
    context.cache["srecord.build_dir"] = buildDir
    context.cache["srecord.install_dir"] = installDir



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Artifacts defintions internally used to refer to intermediate results."""

from enum import Enum
from pathlib import Path

from mlonmcu.setup import utils

# TODO: offer pack/unpack/flatten methods for mlf
# TODO: implement restore methods
# TODO: decide if inheritance based scheme would fit better


class ArtifactFormat(Enum):  # TODO: ArtifactType, ArtifactKind?
    """Enumeration of artifact types."""

    UNKNOWN = 0
    SOURCE = 1
    TEXT = 2
    MLF = 3
    MODEL = 4
    IMAGE = 5
    DATA = 6
    NUMPY = 7
    PARAMS = 8
    JSON = 9  # TODO: how about YAML or more general: DICT?
    PATH = 10  # NOT A DIRECTORY?
    RAW = 11
    BIN = 11
    SHARED_OBJECT = 12  # Here: the parent tar archive
    ARCHIVE = 13


def lookup_artifacts(artifacts, name=None, fmt=None, flags=None, first_only=False):
    """Utility to get a matching artifact for a given set of properties."""
    if isinstance(name, Path):
        name = str(name)
    matches = []
    # Warning: if neither name, fmt nor flags is provided, the first artifact (first_only=True)
    # or all (first_only=False) are returned
    for artifact in artifacts:
        valid = True
        if name is not None and artifact.name != name:
            valid = False
        if fmt is not None and artifact.fmt != fmt:
            valid = False
        if flags is not None and not all(
            flag in artifact.flags for flag in flags
        ):  # A valid artifact may have more flags than specified
            valid = False
        if valid:
            matches.append(artifact)
    if len(matches) > 0 and first_only:
        matches = [matches[0]]
    return matches


class Artifact:
    """Artifact type."""

    def __init__(
        self,
        name,
        content=None,
        path=None,
        data=None,
        raw=None,
        fmt=ArtifactFormat.UNKNOWN,
        flags=None,
        archive=False,
        optional=False,
    ):
        # TODO: Allow to store filenames as well as raw data
        self.name = name
        # TODO: too many attributes...
        self.content = content
        self.path = path
        self.data = data
        self.raw = raw
        self.fmt = fmt
        self.flags = flags if flags is not None else {}
        self.archive = archive
        self.optional = optional
        self.validate()

    def __repr__(self):
        return f"Artifact({self.name}, fmt={self.fmt}, flags={self.flags})"

    @property
    def exported(self):
        """Returns true if the artifact was writtem to disk."""
        return bool(self.path is not None)

    def validate(self):
        """Checker for artifact attributes for the given format."""
        if self.fmt in [ArtifactFormat.TEXT, ArtifactFormat.SOURCE]:
            assert self.content is not None
        elif self.fmt in [ArtifactFormat.RAW, ArtifactFormat.BIN]:
            assert self.raw is not None
        elif self.fmt in [ArtifactFormat.MLF, ArtifactFormat.SHARED_OBJECT, ArtifactFormat.ARCHIVE]:
            assert self.raw is not None
        elif self.fmt in [ArtifactFormat.PATH]:
            assert self.path is not None
        else:
            raise NotImplementedError

    def export(self, dest, extract=False):
        """Export the artifact to a given path (file or directory) and update its path.

        Arguments
        ---------
        dest : str
            Path of the destination.
        extract : bool
            If archive: extract to destination.

        """
        if not isinstance(dest, Path):
            dest = Path(dest)
        if dest.is_dir():
            filename = dest / self.name
        else:
            filename = dest
        if self.fmt in [ArtifactFormat.TEXT, ArtifactFormat.SOURCE]:
            assert not extract, "extract option is only available for ArtifactFormat.MLF"
            with open(filename, "w", encoding="utf-8") as handle:
                handle.write(self.content)
        elif self.fmt in [ArtifactFormat.RAW, ArtifactFormat.BIN]:
            assert not extract, "extract option is only available for ArtifactFormat.MLF"
            with open(filename, "wb") as handle:
                handle.write(self.raw)
        elif self.fmt in [ArtifactFormat.MLF, ArtifactFormat.SHARED_OBJECT, ArtifactFormat.ARCHIVE]:
            with open(filename, "wb") as handle:
                handle.write(self.raw)
            if extract:
                utils.extract(filename, dest)
                # os.remove(filename)
        elif self.fmt in [ArtifactFormat.PATH]:
            assert not extract, "extract option is only available for ArtifactFormat.MLF"
            utils.copy(self.path, filename)
        else:
            raise NotImplementedError
        self.path = filename if self.path is None else self.path

    def print_summary(self):
        """Utility to print information about an artifact to the cmdline."""
        print("Format:", self.fmt)
        print("Optional: ", self.optional)
        if self.fmt in [ArtifactFormat.TEXT, ArtifactFormat.SOURCE]:
            print("Content:")
            print(self.content)
        elif self.fmt in [ArtifactFormat.RAW, ArtifactFormat.BIN]:
            print(f"Data Size: {len(self.raw)}B")
        elif self.fmt in [ArtifactFormat.MLF, ArtifactFormat.SHARED_OBJECT, ArtifactFormat.ARCHIVE]:
            print(f"Archive Size: {len(self.raw)}B")
        elif self.fmt in [ArtifactFormat.PATH]:
            print(f"File Location: {self.path}")
        else:
            raise NotImplementedError
